# ruff: noqa: D205, D401, E501

import asyncio
import dataclasses
import datetime
import uuid

import typing_extensions as typing

from unitelabs.cdk import sila


class InvalidLockIdentifier(Exception):
    """The sent lock identifier is not valid."""


class ServerAlreadyLocked(Exception):
    """The SiLA Server can not be locked because it is already locked."""


class ServerNotLocked(Exception):
    """The SiLA Server can not be unlocked because it is not locked."""


@dataclasses.dataclass
class LockIdentifier(sila.Metadatum, errors=[InvalidLockIdentifier]):
    """The lock identifier has to be sent with every (lock protected) call in order to use the functionality of a locked SiLA Server."""

    lock_identifier: str

    @typing.override
    def intercept(self) -> None:
        if not isinstance(self.feature, LockController):
            raise RuntimeError

        if self.feature.lock is None or self.feature.lock.identifier != self.lock_identifier:
            raise InvalidLockIdentifier

        self.feature.lock.last_usage = sila.datetime.datetime.now()


@dataclasses.dataclass
class Lock:
    """A lock used for unique access to resources."""

    identifier: str = dataclasses.field(default_factory=lambda: str(uuid.uuid4()))
    """The identifier used by the client for unique access."""

    lifetime: datetime.timedelta = dataclasses.field(default_factory=datetime.timedelta)
    """The lifetime of the access token before it expires."""

    last_usage: datetime.datetime = dataclasses.field(default_factory=datetime.datetime.now)
    """Point in time when the access token was used last."""

    @property
    def is_expired(self) -> bool:
        """Whether the lock has expired."""

        return (datetime.datetime.now() - self.lifetime) > self.last_usage


class LockController(sila.Feature):
    """
    This Feature allows a SiLA Client to lock a SiLA Server for exclusive use, preventing other SiLA Clients
    from using the SiLA Server while it is locked. To lock a SiLA Server a Lock Identifier has to be set, using the
    'LockServer' command. This Lock Identifier has to be sent along with every (lock protected)
    request to the SiLA Server in order to use its functionality.

    To send the lock identifier the SiLA Client Meta Data 'LockIdentifier' has to be used.

    When locking a SiLA Server a timeout can be specified that defines the time after which the SiLA Server will
    be automatically unlocked if no request with a valid lock identifier has been received meanwhile.
    After the timeout has expired or after explicit unlock no lock identifier has to be sent any more.
    """

    def __init__(self):
        super().__init__(
            originator="org.silastandard",
            category="core",
            version="2.0",
            metadata=[LockIdentifier],
        )

        self._lock: Lock | None = None
        self._event = asyncio.Event()

    @property
    def lock(self) -> Lock | None:
        """The lock on the SiLA Server, if any."""

        if self._lock and self._lock.is_expired:
            self._event.set()
            self._lock = None

        return self._lock

    @lock.setter
    def lock(self, value: Lock | None) -> None:
        self._event.set()
        self._lock = value

    @sila.ObservableProperty()
    async def subscribe_is_locked(self) -> sila.Stream[bool]:
        """
        Returns true if the SiLA Server is currently locked or false else.

        This property MUST NOT be lock protected, so that any SiLA Client can query the current lock state
        of a SiLA Server.
        """

        while True:
            self._event.clear()
            yield self.lock is not None
            await self._event.wait()

    @sila.UnobservableCommand()
    def lock_server(
        self,
        lock_identifier: str,
        timeout: typing.Annotated[
            int,
            sila.constraints.Unit(label="s", components=[sila.constraints.Unit.Component(unit="Second")]),
        ],
    ) -> None:
        """
        Locks a SiLA Server for exclusive use by setting a lock identifier that has to be sent along with
        any following (lock protected) request as long as the SiLA Server is locked.
        The lock can be reset by issuing the 'Unlock Server' command.

        Args:
          LockIdentifier: The lock identifier that has to be sent along with every (lock protected) request to use the server's functionality.
          Timeout: The time (in seconds) after a SiLA Server is automatically unlocked when no request with a valid lock identifier
            has been received meanwhile. A timeout of zero seconds specifies an infinite time (no timeout).

        Raises:
          ServerAlreadyLocked: The SiLA Server can not be locked because it is already locked.
        """

        if self.lock:
            raise ServerAlreadyLocked

        self.lock = Lock(identifier=lock_identifier, lifetime=sila.datetime.timedelta(seconds=timeout))

    @sila.UnobservableCommand()
    def unlock_server(self, lock_identifier: str) -> None:
        """
        Unlocks a locked SiLA Server. No lock identifier has to be sent for any following calls until
        the server is locked again via the 'Lock Server' command.

        Args:
          LockIdentifier: The lock identifier that has been used to lock the SiLA Server.

        Raises:
          ServerNotLocked: The SiLA Server can not be unlocked because it is not locked.
          InvalidLockIdentifier: The sent lock identifier is not valid.
        """

        lock = self.lock

        if lock is None:
            raise ServerNotLocked

        if lock.identifier != lock_identifier:
            raise InvalidLockIdentifier

        self.lock = None



import re
import unittest.mock

import pytest

from unitelabs.cdk.sila.common.decorator import Decorator
from unitelabs.cdk.sila.common.feature import Feature


class TestAttach:
    def test_should_attach_decorator(self):
        # Create Feature
        mock = unittest.mock.Mock()

        class TestFeature(Feature):
            @Decorator()
            def test(self):
                mock(self)

        feature = TestFeature()

        # Attach decorator
        attached = feature.attach()

        # Assert that the method returns the correct value
        assert attached is True
        assert "Test" in feature._handlers
        assert feature._handlers["Test"]._function.func == TestFeature.test
        assert feature._handlers["Test"]._feature == feature

        feature.test()
        mock.assert_called_once_with(feature)

    def test_should_attach_multiple_decorators(self):
        # Create Feature
        mock = unittest.mock.Mock()

        class TestFeature(Feature):
            @Decorator()
            def test(self):
                mock(self)

        feature1 = TestFeature()
        feature2 = TestFeature()

        # Attach decorator
        attached1 = feature1.attach()
        attached2 = feature2.attach()

        # Assert that the method returns the correct value
        assert attached1 is True
        assert "Test" in feature1._handlers
        assert feature1._handlers["Test"]._function.func == TestFeature.test
        assert feature1._handlers["Test"]._feature == feature1

        assert attached2 is True
        assert "Test" in feature2._handlers
        assert feature2._handlers["Test"]._function.func == TestFeature.test
        assert feature2._handlers["Test"]._feature == feature2

        feature1.test()
        mock.assert_called_once_with(feature1)

        mock.reset_mock()

        feature2.test()
        mock.assert_called_once_with(feature2)

    def test_should_raise_on_duplicate_handler(self):
        # Create Feature
        mock = unittest.mock.Mock()

        class TestFeature(Feature):
            @Decorator(name="Test")
            def test1(self):
                mock(self)

            @Decorator(name="Test")
            def test2(self):
                mock(self)

        feature = TestFeature()

        # Attach decorator
        with pytest.warns(
            UserWarning,
            match=re.escape(
                "Duplicate handler identifier 'Test' detected for feature 'TestFeature'. "
                "Existing: TestFeature.test1. New: TestFeature.test2 (will override). "
                "To avoid unintended overrides, set a unique 'identifier' in your decorator "
                "or rename one of the methods."
            ),
        ):
            feature.attach()



import json
import logging
import logging.config
import pathlib

import deprecated
import ruamel.yaml


@deprecated.deprecated(
    version="0.2.8",
    reason=(
        "We now globally configure logging instead of configuring the individual logger instance,"
        " therefore replace `create_logger` with `logging.getLogger`."
    ),
)
def create_logger(name: str = __package__, level: int = logging.INFO) -> logging.Logger:
    """Get the app's logger and configure it if needed."""

    logger = logging.getLogger(name)
    handlers = [handler for handler in logger.handlers if not isinstance(handler, logging.NullHandler)]

    if not handlers:
        formatter = logging.Formatter("{asctime} [{levelname!s:<8}] {message!s}", style="{")

        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)

        logger.setLevel(level)
        logger.addHandler(stream_handler)

    return logger


def configure_logging(config: None | pathlib.Path | str | dict = None, log_level: int | None = None) -> None:
    """
    Configure logging with the given config or provide a file containing the config.

    Args:
      config: Either a path containing the config or the config itself.
        See https://docs.python.org/3/library/logging.config.html#configuration-dictionary-schema
      log_level: Override the root logger's log level.
    """

    if config is None:
        # only create a basic config if no other config is provided
        logging.basicConfig(format="{asctime} [{levelname!s:<8}] {message!s}", style="{")

    if isinstance(config, (pathlib.Path, str)):
        config = pathlib.Path(config)

        if not config.exists():
            msg = f"Provided configuration file {config} does not exist."
            raise ValueError(msg)

        if config.suffix == ".json":
            config = json.loads(config.read_text(encoding="utf-8"))

        elif config.suffix in (".yaml", ".yml"):
            yaml = ruamel.yaml.YAML()
            config = yaml.load(config.read_text(encoding="utf-8"))

    if isinstance(config, dict):
        logging.config.dictConfig(config)

    elif isinstance(config, pathlib.Path):  # .ini files
        logging.config.fileConfig(config)

    if log_level is not None:
        logging.getLogger().setLevel(log_level)



import dataclasses
import inspect
import warnings
import weakref

import typing_extensions as typing
from sila.server import DataType, Structure

from ..utils import parse_docstring, to_display_name, to_identifier

if typing.TYPE_CHECKING:
    from ..common import Feature


@dataclasses.dataclass
class Dataclass:
    """Base class for dataclass based SiLA annotations."""

    _identifier: typing.ClassVar[str] = ""
    _name: typing.ClassVar[str] = ""
    _description: typing.ClassVar[str] = ""

    def __init_subclass__(
        cls,
        /,
        *,
        identifier: str | None = None,
        display_name: str | None = None,
        name: str | None = None,
    ) -> None:
        super().__init_subclass__()

        if display_name is not None:
            msg = "Using `display_name` is deprecated, use `name` instead."
            warnings.warn(msg, stacklevel=2)
            name = display_name

        cls._name = name or to_display_name(cls.__name__)
        cls._identifier = identifier or to_identifier(cls._name)
        cls._description = inspect.getdoc(cls)
        cls._feature: Feature | None = None

    @classmethod
    def attach(cls, feature: "Feature") -> None:
        """
        Create and attach a dataclass to the `feature`.

        Args:
          feature: The `Feature` to which the dataclass will be attached.
        """

        cls._feature = weakref.proxy(feature)

    @classmethod
    def _infer_data_type(cls, feature: "Feature") -> type[DataType]:
        docstring = parse_docstring(cls, feature=feature)
        cls._description = docstring.description

        if len(docstring.parameters) == 0:
            msg = (
                f"Could not detect any fields on '{cls._identifier}'. "
                "Did you forget to annotated your data type definition with `@dataclasses.dataclass`?"
            )
            raise ValueError(msg)
        if (
            len(docstring.parameters) == 1
            and (item := next(iter(docstring.parameters.values()), None))
            and item.identifier == cls._identifier
        ):
            data_type = item.data_type
        else:
            data_type = Structure.create(
                elements=docstring.parameters, name=cls.__name__, description=docstring.description
            )

        return data_type



import re
import unicodedata
import warnings


def to_display_name(value: str, sep: str = "_") -> str:
    """
    Convert a class name to a SiLA display name.

    Example:
      >>> to_display_name("MyClassName")
      "My Class Name"

      >>> to_display_name("my_variable_name")
      "My Variable Name"
    """

    value = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", value)
    value = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", value)
    value = " ".join(x[0].upper() + x[1:] for x in (value or "").split(sep) if x)

    if len(value) > 255:
        value = value[:255]
        warnings.warn(f"Display '{value}' name is too long and will be truncated.", stacklevel=5)

    return value


def to_identifier(value: str) -> str:
    """
    Convert a SiLA display name to a SiLA identifier.

    Example:
      >>> to_identifier("My Class Name")
      "MyClassName"

      >>> to_identifier("My Variable Name")
      "MyVariableName"
    """

    value = unicodedata.normalize("NFKD", value).encode("ascii", "ignore").decode("ascii")
    value = re.sub(r"[^a-zA-Z0-9]", "", value)

    if not value:
        return ""

    if not value[0].isalpha():
        value = "I" + value

    if not value[0].isupper():
        value = value[0].upper() + value[1:]

    return value



from .certificate import certificate
from .config import config
from .connector import connector
from .start import TLSConfigurationError

__all__ = [
    "TLSConfigurationError",
    "certificate",
    "config",
    "connector",
]



import dataclasses
import gc
import json
import pathlib
import uuid

import pytest
from ruamel.yaml import YAML

from unitelabs.cdk.config import (
    CloudServerConfig,
    ConfigurationError,
    ConnectorBaseConfig,
    SiLAServerConfig,
    UnsupportedConfigFiletype,
    get_connector_config,
)
from unitelabs.cdk.config.config import SerializableDict

BASE_CONFIG_DEFAULT: SerializableDict = {
    "sila_server": {
        "hostname": "0.0.0.0",
        "port": 0,
        "tls": False,
        "require_client_auth": False,
        "root_certificates": None,
        "certificate_chain": None,
        "private_key": None,
        "options": {},
        "name": "SiLA Server",
        "type": "ExampleServer",
        "description": "",
        "version": "0.1",
        "vendor_url": "https://sila-standard.com",
        "default_lifetime": 3600,
    },
    "discovery": {"network_interfaces": [], "ip_version": "ipv4"},
    "cloud_server_endpoint": {
        "hostname": "localhost",
        "port": 50001,
        "tls": True,
        "root_certificates": None,
        "certificate_chain": None,
        "private_key": None,
        "reconnect_delay": 10000,
        "options": {},
    },
    "logging": None,
}
CERT = """
-----BEGIN CERTIFICATE-----
MIIC/jCCAeagAwIBAgIURXGhJmigNdNvgR3wMdlsv43VhCEwDQYJKoZIhvcNAQEL
BQAwEDEOMAwGA1UEAwwFU2lMQTIwHhcNMjQxMDA5MTMwNDU1WhcNMjUxMDA5MTMw
NDU1WjAQMQ4wDAYDVQQDDAVTaUxBMjCCASIwDQYJKoZIhvcNAQEBBQADggEPADCC
AQoCggEBAMCSToXHEzV76zHO3pATJ+M3zRqnbZ9KwyGJzWCA7jmZZWkCligW7QgQ
COf8AdOcf5eZawop63HeDDqkuQtAAKwOUiVjLgoPXpu9l9lxDSBo1XfquTIvgNGY
wN7i2W9zQQ0U78iBJ7+xcEbkf9m/s5yyKXGOds1apBVx184Qb2MKnGc2mK6WkRf3
uIBBj/o3JzNlEu040zIok/A/DtRimgoxOipjzbLYRq5xLtod1tpyw1lsQGctfCNi
27bYApcA2UUpy0PG8BUMbT2jkwLTkruKrfx9x4tcHL3PH9s0oXWzijKFw8seAx15
LIhoyjonygZeoKaEAkT2Rv+WzGeUDRsCAwEAAaNQME4wGgYDVR0RBBMwEYIJbG9j
YWxob3N0hwR/AAABMDAGCCsGAQQBg8lXBCQ1YThmZGI3MS1lZTIzLTQ1MmMtODM1
Yy1lMjI0ZGY3YmU5NTAwDQYJKoZIhvcNAQELBQADggEBAF9NbU6GItVjepD1w4UA
2V4UUfcrKdk9xO0FVEF/lRyPoRaoKzO08k5AJ47tmjqqqpijzYXspKta+UEMiAUE
qoCRhgLCvtrAIp7nhsCNw7fyK+i+866q1dyB2VogDbLwxyAj4nxd4o2flWNPaTqr
XgQ9xGuww3Izngr+MGFKsE9CJ7b5emq67lfGQ8UkwvfU2XhcPWPJAyHi2kQPVXFk
UmmzQoqnqELfSNpnd5zUrgdVVtABHZ+rsCJpldU0yO5OQ4pfm+f56VmS2Da2vDZt
GkEfLq/po/VTGOB6bZ1oV+kR+ZHbRpCh0mAQ+OvuUuWOAmuQ30m26nA+w7+VyLux
ht4=
-----END CERTIFICATE-----

"""
KEY = """
-----BEGIN RSA PRIVATE KEY-----
MIIEpAIBAAKCAQEAwJJOhccTNXvrMc7ekBMn4zfNGqdtn0rDIYnNYIDuOZllaQKW
KBbtCBAI5/wB05x/l5lrCinrcd4MOqS5C0AArA5SJWMuCg9em72X2XENIGjVd+q5
Mi+A0ZjA3uLZb3NBDRTvyIEnv7FwRuR/2b+znLIpcY52zVqkFXHXzhBvYwqcZzaY
rpaRF/e4gEGP+jcnM2US7TjTMiiT8D8O1GKaCjE6KmPNsthGrnEu2h3W2nLDWWxA
Zy18I2LbttgClwDZRSnLQ8bwFQxtPaOTAtOSu4qt/H3Hi1wcvc8f2zShdbOKMoXD
yx4DHXksiGjKOifKBl6gpoQCRPZG/5bMZ5QNGwIDAQABAoIBAE023PFbH2Kkq2uv
TSJr6+R5rW3wkE38xj0eahE14U+LKFRwyxCMEMLY2xlZvMnCyI5a38aVhGiF5lVl
UyUlpp9Wpq2DFSTHgOHlpYt0fxTttBp/LX7n+TkRjNRSFWlQx1adfH/i+bMtTJ3A
ZVtEOJqt/VwhCZXRsFVA7o0bne4R0VZZW/5eDH4nF9JNQsvOo2Qrcx7fRzld4khx
7Y3lbf363RQndONdJ7znPMZMd1RzFWygsjbU8vjcL2rA7bDEHLkqKc6er+RDPTWq
SgnMT2vL/+lzQxuf59JaIBU3nJA02+8jLWjMpRLSbvnNuxV/YN4j9+tzC5ChPm/4
+3HnwBECgYEA/0EtN1hl8SbwEAz+hoVB6pOhjIPOBrKqT/1xRLyPPGt5zlM3vFmc
o60LV5hcWI7dajd4TRc6mYZEYpMkWPY9FQKSANH1WmHwp9rRUwDyTCn6xvza5dk/
cV1ZnvMMjwfKi531NhZQerA4Fg/3CoiToBXIWRBl4+Vhh2Tx1mcUuA8CgYEAwSJE
+nbESRR5rbKEC1Yezc+RuxzhVqIWApOVZx29O6VLMmlL+O1ZiK6Z08YUNNno6otc
Lnhi3I5bv2O5CS4ubUpqtwLuOKKJPLdx8ZBh5aJ/1LVfEmqih58FxzYiZ3+Hfsjy
PaqSwEDp3CcShZ6Tzz8eaeJFaG9k56a659QU7jUCgYAaG2FzirAKhTAChEG4IoMG
agkY5RY6ayWuPr7KB/sic9+mca5+ri+uMfG6CNRRHnOY/IlqYRjWQPxXlLMgAjdn
IbcrLE5K6z+A+4lzUuJ1VcnXdl8xKRIrFyAmeLdtHZ/ivcopuQiMM9/YqdNbmXJ3
6iJusZWqRHjAL1vo0Ow2kwKBgQCNhKTyupA91IkMpCBphiNwP8bTSug7aO2j2azC
MGJ3EDm3qLyInLLcmsQRD7XCvGIVaySS0JfwcUf9R/9QIMzYPI1RqQ4R5deV6/3M
OjXh5F6y6GvPvN93bSj4vkwbdrE8T9ZhJVn/EhHKxb6mtnoshF2uzKR7UBSqQdv2
/8qOeQKBgQDdWqFawsUcMZmq3faoipofg0Gl8pZiKYjVOV5QBz8z2mxxUuHGB3VQ
17kBvR3rnhbyhj/kS0rq7mKib+8K9WjKeZr/ypr1oiOSXKPm5UqZTctFcAqkvgyE
Sz0JRTsDjVBHrdnbVUF6QNh+hqTkqYGMu2RcArnvmMdnQ5D1jMfe0A==
-----END RSA PRIVATE KEY-----

"""

yaml_config_w_comments = """sila_server:
  hostname: 0.0.0.0
  port: 0
  tls: false # comment about tls
  require_client_auth: false
  root_certificates:
  certificate_chain:
  private_key:
  options: {}
  uuid: 3264bf7c-45d0-4fb5-9da4-d4fa53eedb6c
  name: Server Name
  type: ExampleServer
  description: Multi-line description that can span multiple lines
    if needed and still survive the round trip. This should survive
    as well.
  version: 0.1.6_rc1
  vendor_url: https://unitelabs.io/
discovery:
  network_interfaces: []
  ip_version: ipv4
cloud_server_endpoint:  # cloud server comment
  hostname: localhost
  port: 50001
  tls: false
  root_certificates:
  certificate_chain:
  private_key:
  reconnect_delay: 10000.0
  options: {}
logging:
"""


@pytest.fixture()
def config_file(tmp_path) -> pathlib.Path:
    config_path = tmp_path / "config.yaml"
    with config_path.open("w") as f:
        f.write(yaml_config_w_comments)
    yield config_path


@pytest.fixture
def cleanup():
    """Garbage collect derived Config classes to ensure `get_connector_config` finds only one subclass."""
    yield
    gc.collect()


def normalize_whitespace(s: str) -> str:
    """Remove extra whitespace from a string, while preserving comments."""
    return "\n".join("".join(map(str.rstrip, line.split("#"))) for line in s.splitlines()).rstrip()


class TestInnerTypes:
    def test_sila_server_conversion(self):
        default = ConnectorBaseConfig().to_dict()
        instance = ConnectorBaseConfig(**default)
        assert isinstance(instance.sila_server, SiLAServerConfig)

    @pytest.mark.parametrize(
        "path,content,key", [("cert.pem", CERT, "certificate_chain"), ("key.pem", KEY, "private_key")]
    )
    def test_sila_server_cert_key_conversion(self, tmp_path, path, content, key):
        path = tmp_path / path
        path.touch()
        path.write_bytes(content.encode("ascii"))

        default = ConnectorBaseConfig().to_dict()
        default["sila_server"][key] = path

        config = ConnectorBaseConfig(**default)
        assert isinstance(getattr(config.sila_server, key), bytes)
        assert getattr(config.sila_server, key).decode("ascii") == content

    def test_cloud_server_endpoint_conversion(self):
        default = ConnectorBaseConfig().to_dict()
        instance = ConnectorBaseConfig(**default)
        assert isinstance(instance.cloud_server_endpoint, CloudServerConfig)

    @pytest.mark.parametrize(
        "path,content,key", [("cert.pem", CERT, "certificate_chain"), ("key.pem", KEY, "private_key")]
    )
    def test_cloud_server_endpoint_cert_key_conversion(self, tmp_path, path, content, key):
        path = tmp_path / path
        path.touch()
        path.write_bytes(content.encode("ascii"))

        default = ConnectorBaseConfig().to_dict()
        default["cloud_server_endpoint"][key] = path

        config = ConnectorBaseConfig(**default)
        assert isinstance(getattr(config.cloud_server_endpoint, key), bytes)
        assert getattr(config.cloud_server_endpoint, key).decode("ascii") == content

    @pytest.mark.parametrize("key,name", [("certificate_chain", "Certificate chain"), ("private_key", "Private key")])
    def test_sila_server_cert_key_file_not_found(self, key, name):
        default = ConnectorBaseConfig().to_dict()
        path = pathlib.Path("path.json")
        default.update({"sila_server": {f"{key}": path}})
        with pytest.raises(FileNotFoundError, match=rf"File at path '{path.resolve()}' not found."):
            ConnectorBaseConfig(**default)

    @pytest.mark.parametrize("key,name", [("certificate_chain", "Certificate chain"), ("private_key", "Private key")])
    def test_cloud_server_endpoint_cert_key_file_not_found(self, key, name):
        default = ConnectorBaseConfig().to_dict()
        path = pathlib.Path("path.json")
        default.update({"cloud_server_endpoint": {f"{key}": path}})

        with pytest.raises(FileNotFoundError, match=rf"File at path '{path.resolve()}' not found."):
            ConnectorBaseConfig(**default)

    def test_should_not_allow_neither_sila_server_nor_cloud_server(self):
        default = ConnectorBaseConfig().to_dict()
        default["sila_server"] = None
        default["cloud_server_endpoint"] = None

        with pytest.raises(
            ConfigurationError,
            match=r"At least one of 'sila_server' or 'cloud_server_endpoint' must be configured.",
        ):
            ConnectorBaseConfig(**default)

    @pytest.mark.parametrize(
        "hostname,raises",
        [
            pytest.param("http://localhost.dev", True, id="invalid-http"),
            pytest.param("https://localhost.dev", True, id="invalid-https"),
            pytest.param("localhost.dev", False, id="valid"),
        ],
    )
    def test_should_raise_on_invalid_hostname(self, hostname: str, raises: bool):
        default = ConnectorBaseConfig().to_dict()
        default["cloud_server_endpoint"]["hostname"] = hostname
        if raises:
            with pytest.raises(
                ConfigurationError,
                match=r"Hostname must not contain 'http://' or 'https://'.",
            ):
                ConnectorBaseConfig.validate(default)
        else:
            ConnectorBaseConfig.validate(default)


class TestSchema:
    def test_schema(self, cleanup):
        @dataclasses.dataclass
        class ExampleConfig(ConnectorBaseConfig):
            complex: bool = False
            """Whether or not the complex configuration is enabled."""
            complex_name: str | None = ""
            """The name of the complex configuration."""

        schema = ExampleConfig.schema()

        assert schema


class TestFromDict:
    def test_type_conversion(self):
        default = ConnectorBaseConfig().to_dict()
        instance = ConnectorBaseConfig.from_dict(default)
        assert isinstance(instance.sila_server, SiLAServerConfig)
        assert isinstance(instance.cloud_server_endpoint, CloudServerConfig)

    def test_should_allow_arbitrary_fields(self):
        default = ConnectorBaseConfig().to_dict()
        default["this"] = "that"
        instance = ConnectorBaseConfig.from_dict(default)
        assert instance.this == "that"

    def test_should_validate_known_fields(self):
        default: SerializableDict = ConnectorBaseConfig().to_dict()
        default["sila_server"]["port"] = "invalid"
        with pytest.raises(
            ConfigurationError,
            match=r"Invalid configuration for ConnectorBaseConfig: "
            "1 validation error for ConnectorBaseConfig\n"
            "sila_server.port\n  "
            "Input should be a valid integer, unable to parse string as an integer",
        ):
            ConnectorBaseConfig.from_dict(default)


class TestLoad:
    def test_should_load_yaml_file(self, tmp_path):
        config_file_path = tmp_path / "config.yaml"
        config = BASE_CONFIG_DEFAULT.copy()
        config["sila_server"]["uuid"] = str(uuid.uuid4())
        with config_file_path.open("w") as f:
            YAML().dump(config, f)

        instance = ConnectorBaseConfig.load(config_file_path)
        data = dataclasses.asdict(instance)
        assert data == config

    def test_should_load_json_file(self, tmp_path):
        config_file_path = tmp_path / "config.json"
        config = BASE_CONFIG_DEFAULT.copy()
        config["sila_server"]["uuid"] = str(uuid.uuid4())
        with config_file_path.open("w") as f:
            json.dump(config, f)

        instance = ConnectorBaseConfig.load(config_file_path)
        data = dataclasses.asdict(instance)
        assert data == config


class TestDump:
    def test_should_dump_yaml_file(self, tmp_path):
        config_file_path = tmp_path / "config.yaml"
        config = BASE_CONFIG_DEFAULT.copy()
        config["sila_server"]["uuid"] = str(uuid.uuid4())

        ConnectorBaseConfig.from_dict(config).dump(config_file_path)

        with config_file_path.open("r") as f:
            data = YAML().load(f)

        assert data == config

    def test_should_dump_json_file(self, tmp_path):
        config_file_path = tmp_path / "config.json"
        config = BASE_CONFIG_DEFAULT.copy()
        config["sila_server"]["uuid"] = str(uuid.uuid4())

        ConnectorBaseConfig.from_dict(config).dump(config_file_path)

        with config_file_path.open("r") as f:
            data = json.load(f)

        assert data == config

    @pytest.mark.parametrize("ext", ["toml", "ini"])
    def test_should_raise_unsupported_filetype(self, tmp_path, ext):
        config_file_path = tmp_path / f"config.{ext}"
        config_file_path.touch()

        config = BASE_CONFIG_DEFAULT.copy()
        config["sila_server"]["uuid"] = str(uuid.uuid4())

        instance = ConnectorBaseConfig.from_dict(config)
        with pytest.raises(
            UnsupportedConfigFiletype,
            match=f"Cannot write file to {config_file_path.resolve()}. Only yaml and json filetypes are supported.",
        ):
            instance.dump(config_file_path)

    def test_yaml_comment_preservation(self, config_file, tmp_path):
        config = ConnectorBaseConfig.load(config_file)
        test_path = tmp_path / "dump.yaml"
        config.dump(test_path)

        round_trip_data: str = test_path.read_text()

        # Compare while ignoring trailing whitespace differences per line

        assert normalize_whitespace(round_trip_data) == normalize_whitespace(yaml_config_w_comments)

    def test_should_preserve_comment_on_updated_value_line(self, config_file, tmp_path):
        config = ConnectorBaseConfig.load(config_file)
        config.sila_server.tls = True
        out_path = tmp_path / "out.yaml"
        config.dump(out_path)

        assert ConnectorBaseConfig.load(out_path).sila_server.tls
        assert normalize_whitespace(out_path.read_text()) == normalize_whitespace(
            yaml_config_w_comments.replace("tls: false # comment about tls", "tls: true # comment about tls")
        )


class TestGetConnectorConfig:
    def test_should_return_base_if_no_derived_config_exists(self):
        assert get_connector_config() == ConnectorBaseConfig

    def test_should_find_derived_config(self, cleanup):
        @dataclasses.dataclass
        class ExampleConfig(ConnectorBaseConfig):
            value: bool = False

        assert get_connector_config() == ExampleConfig

    def test_should_raise_for_multiple_configs(self, cleanup):
        @dataclasses.dataclass
        class ExampleConfig(ConnectorBaseConfig):
            simple: bool = False

        @dataclasses.dataclass
        class SecondaryExampleConfig(ConnectorBaseConfig):
            simple: bool = True

        with pytest.raises(
            ConfigurationError,
            match=r"Please ensure only one subclass of ConnectorBaseConfig exists.",
        ):
            get_connector_config()



import asyncio
import logging
import pathlib

import click

from .. import utils
from ..config import get_connector_config, read_config_file
from ..config.config import SerializableDict
from ..logging import configure_logging
from ..main import load, run


class TLSConfigurationError(Exception):
    """TLS Configuration is invalid."""


@click.command()
@click.option(
    "--app",
    type=str,
    metavar="IMPORT",
    default=utils.find_factory,
    show_default=False,
    help="The application factory function to load, in the form 'module:name'.",
    envvar="UNITELABS_CDK_APP",
)
@click.option(
    "-cfg",
    "--config-path",
    type=click.Path(path_type=pathlib.Path),
    default=None,
    help="Path to the configuration file.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Increase the verbosity of the default logger. Use a custom log-config for fine grained handling.",
)
@utils.coroutine
async def dev(app: str, config_path: pathlib.Path | None, verbose: int) -> None:
    """Application Entrypoint."""
    import watchfiles

    loaded_app = await load(app)

    config = None
    try:
        _ = read_config_file(config_path)
    except FileNotFoundError:
        click.echo("No config file was found or provided, creating a default configuration.")
        config_path = None
        config = get_connector_config()().to_dict()

    log_level = logging.ERROR - verbose * 10 if verbose else None
    configure_logging(log_level=log_level)

    async def callback(changes: set[tuple[watchfiles.Change, str]]) -> None:
        """Receive file changes."""

        logger = logging.getLogger("Watcher")
        logger.info("Detected file change: %s", changes.pop()[1])

    await watchfiles.arun_process(
        ".", target=process, args=(loaded_app, config_path, config, verbose), callback=callback
    )


def process(app: str, config_path: pathlib.Path | None, config: SerializableDict | None, verbose: int) -> None:
    """Run the connector in a separate process."""

    config = config or read_config_file(config_path)

    log_level = logging.ERROR - verbose * 10 if verbose else None
    log_config = config.get("logging", None)
    configure_logging(log_config, log_level=log_level)

    asyncio.run(run(app, config))



import asyncio
import functools
import unittest.mock
import weakref

import pytest

from unitelabs.cdk.subscriptions import Publisher, Subject

ONE_OP_TIME = 0.01


def first_pipe(x: str) -> str:
    return x.upper()


def second_pipe(x: str) -> dict[str, str]:
    return {"value": x}


class TestSource:
    async def test_should_allow_coroutine_as_source(self):
        msg = "value"

        async def source() -> str:
            return msg

        pub = Publisher[str](source=source, interval=ONE_OP_TIME)
        sub = pub.subscribe()
        assert await sub.get() == msg

        pub.unsubscribe(sub)

    async def test_should_allow_callable_as_source(self):
        msg = "value"

        def source() -> str:
            return msg

        pub = Publisher[str](source=source, interval=ONE_OP_TIME)
        sub = pub.subscribe()
        assert await sub.get() == msg

        pub.unsubscribe(sub)

    async def test_should_allow_partial_coroutine_as_source(self):
        msg = "value"

        async def source(msg: str) -> str:
            return msg

        pub = Publisher[str](source=functools.partial(source, msg), interval=ONE_OP_TIME)
        sub = pub.subscribe()
        assert await sub.get() == msg

        pub.unsubscribe(sub)

    async def test_should_allow_partial_callable_as_source(self):
        msg = "value"

        def source(msg: str) -> str:
            return msg

        pub = Publisher[str](source=functools.partial(source, msg), interval=ONE_OP_TIME)
        sub = pub.subscribe()
        assert await sub.get() == msg

        pub.unsubscribe(sub)

    async def test_should_type_check_source(self):
        # Requires visual check
        def source() -> str:
            return "value"

        pub = Publisher[int](source=source, interval=ONE_OP_TIME)
        sub = pub.subscribe()
        assert await sub.get() == "value"

        pub.unsubscribe(sub)


class TestInit:
    async def test_should_allow_pipe_function(self):
        def pipe_func(x: str) -> str:
            return x.upper()

        pub = Publisher[str](source=lambda: "value", interval=ONE_OP_TIME, pipe=pipe_func)
        sub = pub.subscribe()

        assert await sub.get() == "VALUE"

        pub.unsubscribe(sub)


class TestSubscribe:
    async def test_should_start_polling_after_add(self):
        mock = unittest.mock.Mock(side_effect=[f"value {i}" for i in range(10)])
        pub = Publisher[str](maxsize=10, source=mock, interval=ONE_OP_TIME)
        pub._set = unittest.mock.Mock(wraps=pub._set)

        assert not pub._update_task
        mock.assert_not_called()

        sub = pub.subscribe()
        pub._set.assert_called_once()
        await asyncio.sleep(ONE_OP_TIME)

        assert sub.size >= 1
        mock.assert_called()
        assert pub._update_task

        pub.unsubscribe(sub)

    async def test_should_not_start_polling_again_if_other_subscribers(self):
        pub = Publisher[str](maxsize=10, source=lambda: "value", interval=ONE_OP_TIME)
        pub._set = unittest.mock.Mock(wraps=pub._set)
        assert not pub._update_task

        sub = pub.subscribe()
        assert pub._update_task
        pub._set.assert_called_once()
        pub._set.reset_mock()

        sub2 = pub.subscribe()
        pub._set.assert_not_called()

        for s in [sub, sub2]:
            pub.unsubscribe(s)

    async def test_should_add_current_value_to_new_subscription(self):
        mock = unittest.mock.Mock(side_effect=[f"value {i}" for i in range(10)])
        pub = Publisher[str](maxsize=10, source=mock, interval=ONE_OP_TIME)

        value = "value"
        pub.update(value)

        sub1 = pub.subscribe()
        assert await sub1.get() == value
        mock.assert_not_called()

        sub2 = pub.subscribe()
        assert await sub2.get() == value
        mock.assert_not_called()

        pub.unsubscribe(sub1)
        pub.unsubscribe(sub2)


class TestUnsubscribe:
    async def test_should_cancel_update_task_if_no_subscribers(self):
        # create a data generator for the publisher
        x = 0

        async def get_next_value() -> str:
            nonlocal x
            x += 1
            return f"update {x}"

        pub = Publisher[str](maxsize=10, source=get_next_value, interval=ONE_OP_TIME)
        assert not pub._update_task

        # create subscription and let it run for a while
        sub = pub.subscribe()
        iterations = 5
        await asyncio.sleep(ONE_OP_TIME * iterations)

        # check that internals are set and queue is being populated
        assert sub.size >= iterations
        assert pub._update_task

        # save a reference to the task and remove the subscription
        task = pub._update_task
        pub.unsubscribe(sub)

        # check that internals from source are cleared
        assert not pub._update_task

        # give the task some to time to be gracefully cancelled
        await asyncio.sleep(0.01)
        assert task.cancelled()

    async def test_should_not_cancel_update_task_if_other_subscribers(self):
        # create a data generator for the publisher
        x = 0

        async def get_next_value() -> str:
            nonlocal x
            x += 1
            return f"update {x}"

        pub = Publisher[str](maxsize=10, source=get_next_value, interval=ONE_OP_TIME)
        assert not pub._update_task

        # create two subscriptions and let them run for a while
        sub1 = pub.subscribe()
        sub2 = pub.subscribe()
        iterations = 5
        await asyncio.sleep(ONE_OP_TIME * iterations)

        # check that internals are set and queue is being populated
        assert sub1.size >= iterations
        assert sub2.size >= iterations
        assert pub._update_task

        # save a reference to the task and remove one subscription
        task = pub._update_task
        pub.unsubscribe(sub1)

        # check that internals from source are not cleared
        size_before = sub2.size
        await asyncio.sleep(ONE_OP_TIME)
        assert sub2.size >= size_before
        assert pub._update_task == task

        pub.unsubscribe(sub2)
        assert not pub._update_task


class TestPipe:
    async def test_should_type_check_pipes(self):
        # Requires visual check
        def invalid_pipe(x: int) -> str:
            return "value"

        pub = Publisher[str](source=lambda: "value", interval=ONE_OP_TIME)
        pub.pipe(invalid_pipe)

    async def test_should_create_new_publisher(self):
        mock = unittest.mock.Mock(side_effect=[f"update {i}" for i in range(10)])
        pub = Publisher[str](source=mock, interval=ONE_OP_TIME, maxsize=10)

        piped = pub.pipe(first_pipe).pipe(second_pipe)
        assert piped._pipe == second_pipe

        assert piped != pub
        assert isinstance(piped, Subject)

    async def test_should_add_subscription_to_new_publisher_only(self):
        mock = unittest.mock.Mock(side_effect=[f"update {i}" for i in range(10)])
        pub = Publisher[str](source=mock, interval=ONE_OP_TIME, maxsize=10)

        piped = pub.pipe(first_pipe).pipe(second_pipe)
        assert piped._pipe == second_pipe
        sub = piped.subscribe()

        assert sub in piped.subscribers
        assert sub not in pub.subscribers
        assert piped._parent._parent == weakref.proxy(pub)

        # cleanup
        piped.unsubscribe(sub)

    async def test_should_update_piped_subscription(self):
        mock = unittest.mock.Mock(side_effect=[f"update {i}" for i in range(10)])
        pub = Publisher[str](source=mock, interval=ONE_OP_TIME, maxsize=10)

        piped = pub.pipe(first_pipe).pipe(second_pipe)
        piped.update = unittest.mock.Mock(wraps=piped.update)
        sub = piped.subscribe()

        for i in range(10):
            value = await sub.get()
            piped.update.assert_called_with(f"UPDATE {i}")
            assert value == {"value": f"UPDATE {i}"}
            assert mock.call_count == i + 1

        # cleanup
        piped.unsubscribe(sub)
        assert not pub._update_task

    async def test_should_not_remove_update_task_if_parent_publisher_has_subscription(self):
        mock = unittest.mock.Mock(side_effect=[f"update {i}" for i in range(10)])
        pub = Publisher[str](source=mock, interval=ONE_OP_TIME, maxsize=10)
        main_sub = pub.subscribe()

        # create piped subjects
        first_piped = pub.pipe(first_pipe)
        second_piped = first_piped.pipe(second_pipe)

        sub1 = first_piped.subscribe()
        sub2 = second_piped.subscribe()

        # use piped subjects to get values
        for i in range(10):
            if i == 5:
                pub.unsubscribe(sub1)
            if i < 5:
                # sub1 is still subscribed
                assert await sub1.get() == f"UPDATE {i}"
            assert await sub2.get() == {"value": f"UPDATE {i}"}

            assert pub._update_task
            assert mock.call_count == i + 1

        # ensure that publisher's update task is still active after unsubscribing from all piped subjects
        pub.unsubscribe(sub2)
        assert pub._update_task

        pub.unsubscribe(main_sub)
        assert not pub._update_task


class TestFilter:
    async def test_should_filter_values(self):
        mock = unittest.mock.Mock(side_effect=list(range(10)))
        pub = Publisher[int](source=mock, interval=ONE_OP_TIME, maxsize=10)

        filtered_pub = pub.filter(lambda x: x > 2)
        filtered_pub.update = unittest.mock.Mock(wraps=filtered_pub.update)

        sub = pub.subscribe()
        filtered_sub = filtered_pub.subscribe()
        await asyncio.sleep(ONE_OP_TIME * 5)
        assert await sub.get() == 0
        assert await filtered_sub.get() == 3

        for s in [sub, filtered_sub]:
            pub.unsubscribe(s)


class TestSubscription_Get:
    async def test_should_get_new_value(self):
        mock = unittest.mock.Mock(side_effect=[f"update {i}" for i in range(10)])
        pub = Publisher[str](source=mock, interval=ONE_OP_TIME, maxsize=10)

        sub = pub.subscribe()

        for i in range(10):
            value = await sub.get()
            assert value == f"update {i}"
            assert pub.current == f"update {i}"
            assert mock.call_count == i + 1

        # cleanup
        pub.unsubscribe(sub)

    async def test_should_timeout_if_nothing_matches_predicate(self):
        mock = unittest.mock.Mock(side_effect=[f"update {i}" for i in range(10)])
        pub = Publisher[str](source=mock, interval=ONE_OP_TIME, maxsize=10)

        sub = pub.subscribe()
        with pytest.raises(TimeoutError):
            value = await sub.get(lambda x: "value" in x, timeout=0.05)

        # cleanup
        pub.unsubscribe(sub)



import json
import pathlib
import subprocess

import click

DEFAULT_BRANCH = "feature/cruft-update"


class CruftNotConfigured(Exception):
    """Cruft is not configured."""

    def __init__(self):
        msg = "Run `cruft link https://gitlab.com/unitelabs/cdk/connector-factory.git` to configure cruft."
        super().__init__(msg)


class CiCdNotConfigured(Exception):
    """CI/CD required environment variables are not present."""


@click.option(
    "--repo-url",
    envvar="REPO_API_URL",
    show_envvar=True,
    help="The git api url for the connector to be updated.",
)
@click.option(
    "--token",
    envvar="CRUFT_BOT_TOKEN",
    show_envvar=True,
    help="A git access token with API and write_repository permissions.",
)
@click.option(
    "--branch",
    default=DEFAULT_BRANCH,
    type=str,
    required=False,
    help="Explicitly name the git branch. If running without `--ci`, this will create a new branch and switch to it.",
)
@click.option(
    "--ci",
    is_flag=True,
    default=False,
    help="Run the sync as part of CI/CD. This creates an MR with the changes in the specified branch.",
)
@click.option(
    "--dry-run", is_flag=True, default=False, help="Show the changes that would be applied, without applying them."
)
@click.option(
    "--title",
    default="chore: Cruft Update",
    envvar="MR_TITLE",
    show_envvar=True,
    type=str,
    required=False,
    help="The title to use for the merge request created in CI/CD.",
)
@click.command()
def sync(repo_url: str, token: str, branch: str, ci: bool, dry_run: bool, title: str) -> None:
    """
    Update the current connector boilerplate.

    Check for changes to the connector-factory template and apply using `cruft update`.

    In GitLab CI/CD REPO_API_URL should follow the form "${CI_API_V4_URL}/projects/${CI_PROJECT_ID}"
    """

    if ci:
        msg = ""
        if not repo_url:
            msg += "REPO_API_URL must be set. "
        if not token:
            msg += "CRUFT_BOT_TOKEN must be set. "
        if msg:
            msg += "CI/CD cannot run in the current state."
            raise CiCdNotConfigured(msg)

    has_changes = cruft_check()
    if not has_changes:
        click.echo("No changes to apply.")
        return

    if dry_run:
        subprocess.run(["cruft", "diff"], check=True)
        return

    if ci or (branch != DEFAULT_BRANCH and not ci):
        subprocess.run(["git", "switch", "-c", branch], check=True)

    cruft_info = json.loads(pathlib.Path("./.cruft.json").resolve().read_text())
    runner = cruft_info["context"]["cookiecutter"].get("env", "poetry")  # default poetry for older conn-factory
    subprocess.run([runner, "run", "cruft", "update", "--skip-apply-ask"], check=True)

    if ci:
        push_to_gitlab(branch, repo_url, token, title)


def cruft_check() -> bool:
    """
    Run `cruft check` to see if there are any updates to apply.

    Returns:
      Whether or not there are changes to the cruft template that can be applied.

    Raises:
      NotConfigured: If no `.cruft.json` file can be found.
    """

    has_changes = False
    cruft_file = pathlib.Path("./.cruft.json").resolve()

    if not cruft_file.exists():
        raise CruftNotConfigured
    try:
        subprocess.check_output(["cruft", "check"])
    except subprocess.CalledProcessError as e:
        if (
            e.returncode == 1
            and e.output == b"FAILURE: Project's cruft is out of date! Run `cruft update` to clean this mess up.\n"
        ):
            has_changes = True
        else:
            click.echo(e.output)
            raise e
    return has_changes


def push_to_gitlab(branch: str, repo_url: str, auth: str, title: str) -> None:
    """Commit and push changes, and create an MR on git."""
    cruft_info = json.loads(pathlib.Path("./.cruft.json").resolve().read_text())
    commit_sha = cruft_info["commit"][:8]
    subprocess.run(["git", "add", "."], check=True)
    subprocess.run(["git", "commit", "-m", f'"chore: cruft update to {commit_sha}"'], check=True)
    subprocess.run(["git", "push", "--set-upstream", "origin", branch], check=True)

    git_mr_args = [
        "curl",
        "-sL",
        "--header",
        f"PRIVATE-TOKEN: {auth}",
        "--data",
        f"source_branch={branch}",
        "--data",
        "target_branch=main",
        "--data",
        f"title={title}",
        "-X",
        "POST",
        f"{repo_url}/merge_requests",
    ]
    subprocess.run(git_mr_args, check=True)



from sila.framework import ExecutionMode, Handler, Native, errors, identifiers
from sila.server import Server

from sila import datetime

from . import constraints, data_types, utils
from .command import Intermediate, ObservableCommand, Status, UnobservableCommand
from .common import Dataclass, Decorator, DefinedExecutionError, Feature, define_error
from .data_types.custom_data_type import CustomDataType
from .metadata import Metadata, Metadatum
from .property import ObservableProperty, Stream, UnobservableProperty

Any = Native

__all__ = [
    "Any",
    "CustomDataType",
    "Dataclass",
    "Decorator",
    "DefinedExecutionError",
    "ExecutionMode",
    "Feature",
    "Handler",
    "Intermediate",
    "Metadata",
    "Metadatum",
    "ObservableCommand",
    "ObservableProperty",
    "Server",
    "Status",
    "Stream",
    "UnobservableCommand",
    "UnobservableProperty",
    "constraints",
    "data_types",
    "datetime",
    "define_error",
    "errors",
    "identifiers",
    "utils",
]



import asyncio
import collections.abc
import inspect

import typing_extensions as typing

from ..sila.utils import clear_interval, set_interval
from .default import _DEFAULT_VALUE
from .subject import IN, OUT, PipeFunction, Subject

if typing.TYPE_CHECKING:
    from .subject import PipeFunction
    from .subscription import Subscription


class Publisher(typing.Generic[IN, OUT], Subject[IN, OUT]):
    """
    An observable which updates itself by polling a data source.

    Args:
      source: A function or coroutine that will be called at a fixed interval as the data source of the subscription.
      interval: How many seconds to wait between polling calls to `source`.
      maxsize: The maximum number of messages to track in the queue.

    Examples:
      Subscribe to a publisher which will call `method` every 2 seconds:
      >>> publisher = Publisher[str](source=method, interval=2, maxsize=10)
      >>> async for state in publisher.subscribe():
      >>>     yield state
    """

    def __init__(
        self,
        source: typing.Callable[[], collections.abc.Awaitable[IN]] | typing.Callable[[], IN],
        interval: float = 5,
        maxsize: int = 0,
        pipe: PipeFunction[IN, OUT] | None = None,
    ) -> None:
        super().__init__(maxsize=maxsize, pipe=pipe)

        self._update_task: asyncio.Task | None = None
        self._source = source
        self._interval = interval

    @typing.override
    def on_subscribe(self) -> None:
        self._set()

    @typing.override
    def _on_subscribe(self, subscription: "Subscription") -> None:
        super()._on_subscribe(subscription)
        if self.current is not _DEFAULT_VALUE:
            subscription.update(typing.cast("OUT", self.current))

    @typing.override
    def on_unsubscribe(self) -> None:
        self._unset()

    def _set(self) -> None:
        """
        Create a background task to poll the data `source` and update the current value.

        Task will be destroyed when all subscriptions to the `Publisher` are removed.
        """
        if not self._update_task:
            self._update_task = set_interval(self.__self_update, delay=self._interval)

    def _unset(self) -> None:
        """
        Stop the background task that polls the data `source`.

        This is called when all subscriptions to the `Publisher` are removed.
        """
        if self._update_task:
            clear_interval(self._update_task)
            self._update_task = None
            self._value = _DEFAULT_VALUE

    async def __self_update(self) -> None:
        new_value = self._source()
        if inspect.isawaitable(new_value):
            new_value = await new_value
        self.update(typing.cast("IN", new_value))



import typing

from sila import Native, datetime
from unitelabs.cdk import sila


class ParameterConstraintsTest(sila.Feature):
    """
    This is a test feature to generically test constrained parameters.

    A validation Error must be raised if a constraint is not fulfilled.
    """

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    # String type constraints

    @sila.UnobservableCommand()
    async def check_string_constraint_length(
        self, constrained_parameter: typing.Annotated[str, sila.constraints.Length(10)]
    ) -> None:
        """
        Test the Length constraint on the String type.

        Args:
          ConstrainedParameter: A String parameter with a Length constraint of 10
        """

    @sila.UnobservableCommand()
    async def check_string_constraint_minimal_length(
        self, constrained_parameter: typing.Annotated[str, sila.constraints.MinimalLength(5)]
    ) -> None:
        """
        Test the MinimalLength constraint on the String type.

        Args:
          ConstrainedParameter: A String parameter with a MinimalLength constraint of 5
        """

    @sila.UnobservableCommand()
    async def check_string_constraint_maximal_length(
        self, constrained_parameter: typing.Annotated[str, sila.constraints.MaximalLength(20)]
    ) -> None:
        """
        Test the MaximalLength constraint on the String type.

        Args:
          ConstrainedParameter: A String parameter with a MaximalLength constraint of 20
        """

    @sila.UnobservableCommand()
    async def check_string_constraint_minmax_length(
        self,
        constrained_parameter: typing.Annotated[
            str, sila.constraints.MinimalLength(5), sila.constraints.MaximalLength(20)
        ],
    ) -> None:
        """
        Test the combination of MinimalLength and MaximalLength constraints on the String type.

        Args:
          ConstrainedParameter: A String parameter with a MinimalLength constraint of 5
            and a MaximalLength constraint of 20
        """

    @sila.UnobservableCommand()
    async def check_string_constraint_set(
        self,
        constrained_parameter: typing.Annotated[
            str, sila.constraints.Set(["First option", "Second option", "Third option"])
        ],
    ) -> None:
        """
        Test the Set constraint on the String type.

        Args:
          ConstrainedParameter: A String parameter with a Set constraint of
            'First option', 'Second option', 'Third option'
        """

    @sila.UnobservableCommand()
    async def check_string_constraint_pattern(
        self,
        constrained_parameter: typing.Annotated[str, sila.constraints.Pattern(r"[0-9]{2}/[0-9]{2}/[0-9]{4}")],
    ) -> None:
        """
        Test the Pattern constraint on String type.

        Args:
          ConstrainedParameter: A String parameter with a Pattern constraint
            of a date of birth in dd/mm/yyyy format
        """

    @sila.UnobservableCommand()
    async def check_string_constraint_content_type(
        self,
        constrained_parameter: typing.Annotated[str, sila.constraints.ContentType("application", "xml")],
    ) -> None:
        """
        Test the ContentType constraint on the String type.

        Args:
          ConstrainedParameter: A String parameter with a ContentType constraint
            allowing only valid XML document as content
        """

    @sila.UnobservableCommand()
    async def check_string_constraint_fully_qualified_identifier(
        self,
        feature_identifier: typing.Annotated[str, sila.constraints.FullyQualifiedIdentifier("FeatureIdentifier")],
        command_identifier: typing.Annotated[str, sila.constraints.FullyQualifiedIdentifier("CommandIdentifier")],
        command_parameter_identifier: typing.Annotated[
            str, sila.constraints.FullyQualifiedIdentifier("CommandParameterIdentifier")
        ],
        command_response_identifier: typing.Annotated[
            str, sila.constraints.FullyQualifiedIdentifier("CommandResponseIdentifier")
        ],
        intermediate_command_response_identifier: typing.Annotated[
            str, sila.constraints.FullyQualifiedIdentifier("IntermediateCommandResponseIdentifier")
        ],
        execution_error_identifier: typing.Annotated[
            str, sila.constraints.FullyQualifiedIdentifier("DefinedExecutionErrorIdentifier")
        ],
        property_identifier: typing.Annotated[str, sila.constraints.FullyQualifiedIdentifier("PropertyIdentifier")],
        custom_data_type_identifier: typing.Annotated[str, sila.constraints.FullyQualifiedIdentifier("TypeIdentifier")],
        metadata_identifier: typing.Annotated[str, sila.constraints.FullyQualifiedIdentifier("MetadataIdentifier")],
    ) -> None:
        """
        Test the FullyQualifiedIdentifier constraint on the String type.

        Args:
          FeatureIdentifier: Tests that the parameter is a valid feature identifier
          CommandIdentifier: Tests that the parameter is a valid command identifier
          CommandParameterIdentifier: Tests that the parameter is a valid command parameter identifier
          CommandResponseIdentifier: Tests that the parameter is a valid command response identifier
          IntermediateCommandResponseIdentifier: Tests that the parameter is a valid intermediate
            command response identifier
          ExecutionErrorIdentifier: Tests that the parameter is a valid execution error identifier
          PropertyIdentifier: Tests that the parameter is a valid property identifier
          CustomDataTypeIdentifier: Tests that the parameter is a valid data type identifier
          MetadataIdentifier: Tests that the parameter is a valid metadata identifier
        """

    @sila.UnobservableCommand()
    async def check_string_constraint_schema(
        self,
        constrained_parameter: typing.Annotated[
            str,
            sila.constraints.Schema(
                "Xml",
                url="https://gitlab.com/SiLA2/sila_base/-/raw/master/schema/FeatureDefinition.xsd",
            ),
        ],
    ) -> None:
        """
        Test that the parameter is valid for a given Schema constraint.

        Args:
          ConstrainedParameter: A String parameter with a Schema constraint allowing only content that
            is compliant with the SiLA 2 Feature Definition schema
        """

    # Integer type constraints

    @sila.UnobservableCommand()
    async def check_integer_constraint_set(
        self, constrained_parameter: typing.Annotated[int, sila.constraints.Set([1, 2, 3])]
    ) -> None:
        """
        Test the Set constraint on the Integer type.

        Args:
          ConstrainedParameter: An Integer parameter with a Set constraint of 1, 2, 3
        """

    @sila.UnobservableCommand()
    async def check_scientifically_notated_integer_limit(
        self,
        constrained_parameter: typing.Annotated[
            int, sila.constraints.MinimalInclusive(int(-1e3)), sila.constraints.MaximalExclusive(int(1.23e5))
        ],
    ) -> None:
        """
        Test if scientific notation is accepted for checking limits of type Integer.

        Args:
          ConstrainedParameter: An Integer parameter with a MinimalInclusive constraint
            of -1e3 and a MaximalExclusive constraint of 1.23e5
        """

    @sila.UnobservableCommand()
    async def check_integer_constraint_maximal_exclusive(
        self,
        constrained_parameter: typing.Annotated[int, sila.constraints.MaximalExclusive(11)],
    ) -> None:
        """
        Test the MaximalExclusive constraint on the Integer type.

        Args:
          ConstrainedParameter: An Integer parameter with a MaximalExclusive constraint of 11
        """

    @sila.UnobservableCommand()
    async def check_integer_constraint_maximal_inclusive(
        self,
        constrained_parameter: typing.Annotated[int, sila.constraints.MaximalInclusive(10)],
    ) -> None:
        """
        Test the MaximalInclusive constraint on the Integer type.

        Args:
          ConstrainedParameter: An Integer parameter with a MaximalInclusive constraint of 10
        """

    @sila.UnobservableCommand()
    async def check_integer_constraint_minimal_exclusive(
        self,
        constrained_parameter: typing.Annotated[int, sila.constraints.MinimalExclusive(-1)],
    ) -> None:
        """
        Test the MinimalExclusive constraint on the Integer type.

        Args:
          ConstrainedParameter: An Integer parameter with a MinimalExclusive constraint of -1
        """

    @sila.UnobservableCommand()
    async def check_integer_constraint_minimal_inclusive(
        self,
        constrained_parameter: typing.Annotated[int, sila.constraints.MinimalInclusive(0)],
    ) -> None:
        """
        Test the MinimalInclusive constraint on the Integer type.

        Args:
          ConstrainedParameter: An Integer parameter with a MinimalInclusive constraint of 0
        """

    @sila.UnobservableCommand()
    async def check_integer_constraint_minmax(
        self,
        constrained_parameter: typing.Annotated[
            int, sila.constraints.MinimalExclusive(-1), sila.constraints.MaximalInclusive(10)
        ],
    ) -> None:
        """
        Test the combination of MinimalExclusive and MaximalInclusive constraints on Integer type.

        Args:
          ConstrainedParameter: An Integer parameter with a MinimalExclusive constraint of -1 and
            a MaximalInclusive constraint of 10
        """

    @sila.UnobservableCommand()
    async def check_integer_constraint_unit(
        self,
        constrained_parameter: typing.Annotated[
            int, sila.constraints.Unit("mL", [sila.constraints.UnitComponent("Meter", 3)], factor=0.000001)
        ],
    ) -> None:
        """
        Test the Unit constraint on the Integer type.

        Args:
          ConstrainedParameter: An Integer parameter with a Unit constraint for milliliter (mL)
        """

    # Real type constraints

    @sila.UnobservableCommand()
    async def check_real_constraint_set(
        self, constrained_parameter: typing.Annotated[float, sila.constraints.Set([1.11, 2.22, 3.33])]
    ) -> None:
        """
        Test the Set constraint on the Real type.

        Args:
          ConstrainedParameter: A Real parameter with a Set constraint of 1.11, 2.22, 3.33
        """

    @sila.UnobservableCommand()
    async def check_scientifically_notated_real_limit(
        self,
        constrained_parameter: typing.Annotated[
            float, sila.constraints.MinimalInclusive(1.23e-3), sila.constraints.MaximalExclusive(1.23456e2)
        ],
    ) -> None:
        """
        Test if scientific notation is accepted for checking limits of type Real.

        Args:
          ConstrainedParameter: A Real parameter with a MinimalInclusive constraint
            of 1.23e-3 and a MaximalExclusive constraint of 1.23456e2
        """

    @sila.UnobservableCommand()
    async def check_real_constraint_maximal_exclusive(
        self,
        constrained_parameter: typing.Annotated[float, sila.constraints.MaximalExclusive(10.5)],
    ) -> None:
        """
        Test the MaximalExclusive constraint on the Real type.

        Args:
          ConstrainedParameter: A Real parameter with a MaximalExclusive constraint of 10.5
        """

    @sila.UnobservableCommand()
    async def check_real_constraint_maximal_inclusive(
        self,
        constrained_parameter: typing.Annotated[float, sila.constraints.MaximalInclusive(10)],
    ) -> None:
        """
        Test the MaximalInclusive constraint on the Real type.

        Args:
          ConstrainedParameter: A Real parameter with a MaximalInclusive constraint of 10
        """

    @sila.UnobservableCommand()
    async def check_real_constraint_minimal_exclusive(
        self,
        constrained_parameter: typing.Annotated[float, sila.constraints.MinimalExclusive(0)],
    ) -> None:
        """
        Test the MinimalExclusive constraint on the Real type.

        Args:
          ConstrainedParameter: A Real parameter with a MinimalExclusive constraint of 0
        """

    @sila.UnobservableCommand()
    async def check_real_constraint_minimal_inclusive(
        self,
        constrained_parameter: typing.Annotated[float, sila.constraints.MinimalInclusive(-1.5)],
    ) -> None:
        """
        Test the MinimalInclusive constraint on the Real type.

        Args:
          ConstrainedParameter: A Real parameter with a MinimalInclusive constraint of -1.5
        """

    @sila.UnobservableCommand()
    async def check_real_constraint_minmax(
        self,
        constrained_parameter: typing.Annotated[
            float, sila.constraints.MinimalExclusive(0.333), sila.constraints.MaximalInclusive(11.111)
        ],
    ) -> None:
        """
        Test the combination of MinimalExclusive and MaximalInclusive constraints on Real type.

        Args:
          ConstrainedParameter: A Real parameter with a MinimalExclusive constraint of 0.333 and
            a MaximalInclusive constraint of 11.111
        """

    @sila.UnobservableCommand()
    async def check_real_constraint_unit(
        self,
        constrained_parameter: typing.Annotated[
            float, sila.constraints.Unit("°C", [sila.constraints.UnitComponent("Kelvin", 1)], offset=273.15)
        ],
    ) -> None:
        """
        Test the Unit constraint on the Real type.

        Args:
          ConstrainedParameter: A Real parameter with a Unit constraint for degree celsius (°C)
        """

    # Date type constraints

    @sila.UnobservableCommand()
    async def check_date_constraint_set(
        self,
        constrained_parameter: typing.Annotated[
            datetime.date,
            sila.constraints.Set(
                [
                    datetime.date(2022, 9, 10),
                    datetime.date(2022, 9, 11, tzinfo=datetime.timezone(datetime.timedelta(hours=-1, minutes=45))),
                    datetime.date(2022, 9, 12, tzinfo=datetime.timezone(datetime.timedelta(hours=11, minutes=30))),
                ]
            ),
        ],
    ) -> None:
        """
        Test the Set constraint on the Date type.

        Args:
          ConstrainedParameter: A Date parameter with a Set constraint of
            2022-09-10Z, 2022-09-11-00:15, 2022-09-12+11:30
        """

    @sila.UnobservableCommand()
    async def check_date_constraint_maximal_exclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.date,
            sila.constraints.MaximalExclusive(
                datetime.date(2019, 12, 31, tzinfo=datetime.timezone(datetime.timedelta(hours=2, minutes=30)))
            ),
        ],
    ) -> None:
        """
        Test the MaximalExclusive constraint on the Date type.

        Args:
          ConstrainedParameter: A Date parameter with a MaximalExclusive constraint of 2019-12-31+02:30
        """

    @sila.UnobservableCommand()
    async def check_date_constraint_maximal_inclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.date,
            sila.constraints.MaximalInclusive(
                datetime.date(2025, 12, 31, tzinfo=datetime.timezone(datetime.timedelta(hours=-2, minutes=30)))
            ),
        ],
    ) -> None:
        """
        Test the MaximalInclusive constraint on the Date type.

        Args:
          ConstrainedParameter: A Date parameter with a MaximalInclusive constraint of 2025-12-31-01:30
        """

    @sila.UnobservableCommand()
    async def check_date_constraint_minimal_exclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.date,
            sila.constraints.MinimalExclusive(
                datetime.date(2019, 1, 1, tzinfo=datetime.timezone(datetime.timedelta(hours=-1, minutes=0)))
            ),
        ],
    ) -> None:
        """
        Test the MinimalExclusive constraint on the Date type.

        Args:
          ConstrainedParameter: A Date parameter with a MinimalExclusive constraint of 2019-01-01-01:00
        """

    @sila.UnobservableCommand()
    async def check_date_constraint_minimal_inclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.date,
            sila.constraints.MinimalInclusive(
                datetime.date(1789, 7, 14, tzinfo=datetime.timezone(datetime.timedelta(hours=2, minutes=0)))
            ),
        ],
    ) -> None:
        """
        Test the MinimalInclusive constraint on the Date type.

        Args:
          ConstrainedParameter: A Date parameter with a MinimalInclusive constraint of 1789-07-14+02:00
        """

    @sila.UnobservableCommand()
    async def check_date_constraint_minmax(
        self,
        constrained_parameter: typing.Annotated[
            datetime.date,
            sila.constraints.MinimalExclusive(
                datetime.date(1976, 6, 1, tzinfo=datetime.timezone(datetime.timedelta(hours=2, minutes=0)))
            ),
            sila.constraints.MaximalInclusive(
                datetime.date(2076, 6, 10, tzinfo=datetime.timezone(datetime.timedelta(hours=2, minutes=0)))
            ),
        ],
    ) -> None:
        """
        Test the combination of MinimalInclusive and MaximalExclusive constraints on Date type.

        Args:
          ConstrainedParameter: A Date parameter with a MinimalInclusive constraint of 1976-06-10+02:00 and
            a MaximalExclusive constraint of 2076-06-10+02:00
        """

    # Time type constraints

    @sila.UnobservableCommand()
    async def check_time_constraint_set(
        self,
        constrained_parameter: typing.Annotated[
            datetime.time,
            sila.constraints.Set(
                [
                    datetime.time(18, 59, 59, 999000, tzinfo=datetime.timezone(datetime.timedelta(hours=2))),
                    datetime.time(19, 0, 0, 0, tzinfo=datetime.timezone(datetime.timedelta(hours=-1, minutes=45))),
                    datetime.time(19, 59, 0, 123000),
                ]
            ),
        ],
    ) -> None:
        """
        Test the Set constraint on the Time type.

        Args:
          ConstrainedParameter: A Time parameter with a Set constraint of
            18:59:59.999+02:00, 19:00:00.000-00:15, 19:59:00.123Z
        """

    @sila.UnobservableCommand()
    async def check_time_constraint_maximal_exclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.time,
            sila.constraints.MaximalExclusive(
                datetime.time(12, 0, 0, 0, tzinfo=datetime.timezone(datetime.timedelta(hours=0, minutes=15)))
            ),
        ],
    ) -> None:
        """
        Test the MaximalExclusive constraint on the Time type.

        Args:
          ConstrainedParameter: A Time parameter with a MaximalExclusive constraint of 12:00:00.000+00:15
        """

    @sila.UnobservableCommand()
    async def check_time_constraint_maximal_inclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.time,
            sila.constraints.MaximalInclusive(
                datetime.time(18, 59, 59, 987000, tzinfo=datetime.timezone(datetime.timedelta(hours=2)))
            ),
        ],
    ) -> None:
        """
        Test the MaximalInclusive constraint on the Time type.

        Args:
          ConstrainedParameter: A Time parameter with a MaximalInclusive constraint of 18:59:59.987+02:00
        """

    @sila.UnobservableCommand()
    async def check_time_constraint_minimal_exclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.time,
            sila.constraints.MinimalExclusive(
                datetime.time(1, 2, 3, 456000, tzinfo=datetime.timezone(datetime.timedelta(hours=7)))
            ),
        ],
    ) -> None:
        """
        Test the MinimalExclusive constraint on the Time type.

        Args:
          ConstrainedParameter: A Time parameter with a MinimalExclusive constraint of 01:02:03.456+07:00
        """

    @sila.UnobservableCommand()
    async def check_time_constraint_minimal_inclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.time,
            sila.constraints.MinimalInclusive(datetime.time(8, 0, 0, 555000)),
        ],
    ) -> None:
        """
        Test the MinimalInclusive constraint on the Time type.

        Args:
          ConstrainedParameter: A Time parameter with a MinimalInclusive constraint of 08:00.555Z
        """

    @sila.UnobservableCommand()
    async def check_time_constraint_minmax(
        self,
        constrained_parameter: typing.Annotated[
            datetime.time,
            sila.constraints.MinimalExclusive(
                datetime.time(8, 0, 0, 500000, tzinfo=datetime.timezone(datetime.timedelta(hours=-3, minutes=30)))
            ),
            sila.constraints.MaximalInclusive(
                datetime.time(18, 59, 59, 595000, tzinfo=datetime.timezone(datetime.timedelta(hours=2, minutes=30)))
            ),
        ],
    ) -> None:
        """
        Test the combination of MinimalInclusive and MaximalExclusive constraints on Time type.

        Args:
          ConstrainedParameter: A Time parameter with a MinimalExclusive constraint of 08:00:00.500-02:30 and a
            MaximalInclusive constraint of 18:59:59.595+02:30
        """

    # Timestamp type constraints

    @sila.UnobservableCommand()
    async def check_timestamp_constraint_set(
        self,
        constrained_parameter: typing.Annotated[
            datetime.datetime,
            sila.constraints.Set(
                [
                    datetime.datetime(1969, 7, 16, 13, 32, 00, 1000),
                    datetime.datetime(1969, 7, 20, 20, 17, 40),
                    datetime.datetime(1969, 7, 24, 16, 50, 35, 123000),
                ]
            ),
        ],
    ) -> None:
        """
        Test the Set constraint on the Timestamp type.

        Args:
          ConstrainedParameter: A Timestamp parameter with a Set constraint of 1969-07-16T13:32:00.001Z,
            1969-07-20T20:17:40Z, 1969-07-24T16:50:35.123+00:00
        """

    @sila.UnobservableCommand()
    async def check_timestamp_constraint_maximal_exclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.datetime,
            sila.constraints.MaximalExclusive(
                datetime.datetime(2022, 9, 7, 16, 0, 0, 0, tzinfo=datetime.timezone(datetime.timedelta(hours=2)))
            ),
        ],
    ) -> None:
        """
        Test the MaximalExclusive constraint on the Timestamp type.

        Args:
          ConstrainedParameter: A Timestamp parameter with a MaximalExclusive constraint of
            2022-09-07T16:00:00.000+02:00
        """

    @sila.UnobservableCommand()
    async def check_timestamp_constraint_maximal_inclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.datetime,
            sila.constraints.MaximalInclusive(
                datetime.datetime(
                    2025, 12, 31, 13, 59, 59, 999000, tzinfo=datetime.timezone(datetime.timedelta(hours=-2, minutes=30))
                )
            ),
        ],
    ) -> None:
        """
        Test the MaximalInclusive constraint on the Timestamp type.

        Args:
          ConstrainedParameter: A Timestamp parameter with a MaximalInclusive constraint of
            2025-12-31T13:59:59.999-01:30
        """

    @sila.UnobservableCommand()
    async def check_timestamp_constraint_minimal_exclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.datetime,
            sila.constraints.MinimalExclusive(
                datetime.datetime(2019, 1, 1, 0, 0, 0, 0, tzinfo=datetime.timezone(datetime.timedelta(hours=-1)))
            ),
        ],
    ) -> None:
        """
        Test the MinimalExclusive constraint on the Timestamp type.

        Args:
          ConstrainedParameter: A Timestamp parameter with a MinimalExclusive constraint of 2019-01-01T00:00:00-01:00
        """

    @sila.UnobservableCommand()
    async def check_timestamp_constraint_minimal_inclusive(
        self,
        constrained_parameter: typing.Annotated[
            datetime.datetime,
            sila.constraints.MinimalInclusive(
                datetime.datetime(
                    2022, 9, 7, 13, 48, 59, 666000, tzinfo=datetime.timezone(datetime.timedelta(hours=2, minutes=0))
                )
            ),
        ],
    ) -> None:
        """
        Test the MinimalInclusive constraint on the Timestamp type.

        Args:
          ConstrainedParameter: A Timestamp parameter with a MinimalInclusive constraint of
            2022-09-07T13:48:59.666+02:00
        """

    @sila.UnobservableCommand()
    async def check_timestamp_constraint_minmax(
        self,
        constrained_parameter: typing.Annotated[
            datetime.datetime,
            sila.constraints.MinimalExclusive(datetime.datetime(1969, 7, 20, 20, 17, 40, 303000)),
            sila.constraints.MaximalInclusive(datetime.datetime(1972, 12, 11, 19, 54, 57, 59000)),
        ],
    ) -> None:
        """
        Test the combination of MinimalInclusive and MaximalExclusive constraints on Timestamp type.

        Args:
          ConstrainedParameter: A Timestamp parameter with a MinimalInclusive constraint of 1969-07-20T20:17:40.303Z and
            a MaximalExclusive constraint of 1972-12-11T19:54:57.059Z
        """

    # List type constraints

    @sila.UnobservableCommand()
    async def check_list_constraint_element_count(
        self,
        constrained_parameter: typing.Annotated[list[str], sila.constraints.ElementCount(5)],
    ) -> None:
        """
        Test the ElementCount constraint on the List type.

        Args:
          ConstrainedParameter: A String List parameter with an ElementCount constraint of 5
        """

    @sila.UnobservableCommand()
    async def check_list_constraint_minimal_element_count(
        self,
        constrained_parameter: typing.Annotated[list[str], sila.constraints.MinimalElementCount(3)],
    ) -> None:
        """
        Test the MinimalElementCount constraint on the List type.

        Args:
          ConstrainedParameter: A String List parameter with an MinimalElementCount constraint of 3
        """

    @sila.UnobservableCommand()
    async def check_list_constraint_maximal_element_count(
        self,
        constrained_parameter: typing.Annotated[list[str], sila.constraints.MaximalElementCount(5)],
    ) -> None:
        """
        Test the MaximalElementCount constraint on the List type.

        Args:
          ConstrainedParameter: A String List parameter with an MaximalElementCount constraint of 5
        """

    @sila.UnobservableCommand()
    async def check_list_constraint_minmax_element_count(
        self,
        constrained_parameter: typing.Annotated[
            list[str], sila.constraints.MinimalElementCount(2), sila.constraints.MaximalElementCount(5)
        ],
    ) -> None:
        """
        Test the combination of MinimalElementCount and MaximalElementCount constraints on the List type.

        Args:
          ConstrainedParameter: A String List parameter with a MinimalElementCount constraint of 2 and
            a MaximalElementCount constraint of 5
        """

    # Binary type constraints

    @sila.UnobservableCommand()
    async def check_binary_constraint_length(
        self,
        constrained_parameter: typing.Annotated[bytes, sila.constraints.Length(10)],
    ) -> None:
        """
        Test the Length constraint on the Binary type.

        Args:
          ConstrainedParameter: A Binary parameter with a Length constraint of 10
        """

    @sila.UnobservableCommand()
    async def check_binary_constraint_minimal_length(
        self,
        constrained_parameter: typing.Annotated[bytes, sila.constraints.MinimalLength(5)],
    ) -> None:
        """
        Test the MinimalLength constraint on the Binary type.

        Args:
          ConstrainedParameter: A Binary parameter with a MinimalLength constraint of 5
        """

    @sila.UnobservableCommand()
    async def check_binary_constraint_maximal_length(
        self,
        constrained_parameter: typing.Annotated[bytes, sila.constraints.MaximalLength(20)],
    ) -> None:
        """
        Test the MaximalLength constraint on the Binary type.

        Args:
          ConstrainedParameter: A Binary parameter with a MaximalLength constraint of 20
        """

    @sila.UnobservableCommand()
    async def check_binary_constraint_minmax_length(
        self,
        constrained_parameter: typing.Annotated[
            bytes, sila.constraints.MinimalLength(5), sila.constraints.MaximalLength(20)
        ],
    ) -> None:
        """
        Test the combination of MinimalLength and MaximalLength constraints on the Binary type.

        Args:
          ConstrainedParameter: A Binary parameter with a MinimalLength constraint of 5 and
            a MaximalLength constraint of 20
        """

    @sila.UnobservableCommand()
    async def check_binary_constraint_content_type(
        self,
        constrained_parameter: typing.Annotated[bytes, sila.constraints.ContentType("application", "xml")],
    ) -> None:
        """
        Test the ContentType constraint on the Binary type.

        Args:
          ConstrainedParameter: A Binary parameter with a ContentType constraint allowing
            only valid XML document as content
        """

    @sila.UnobservableCommand()
    async def check_binary_constraint_schema(
        self,
        constrained_parameter: typing.Annotated[
            bytes,
            sila.constraints.Schema(
                "Xml",
                url="https://gitlab.com/SiLA2/sila_base/-/raw/master/schema/FeatureDefinition.xsd",
            ),
        ],
    ) -> None:
        """
        Test that the parameter is valid for a given Schema constraint.

        Args:
          ConstrainedParameter: A Binary parameter with a Schema constraint allowing only content that is
            compliant with the SiLA 2 Feature Definition schema
        """

    # Any type constraints

    @sila.UnobservableCommand()
    async def check_allowed_types_constraint(
        self,
        constrained_parameter: typing.Annotated[
            Native,
            sila.constraints.AllowedTypes(
                [sila.data_types.Integer, sila.data_types.List.create(sila.data_types.Integer)]
            ),
        ],
    ) -> None:
        """
        Test the AllowedTypes constraint on the Any type.

        Args:
          ConstrainedParameter: An Any type parameter with an AllowedTypes constraint that only
            allows an integer or a list of integers
        """

    @sila.UnobservableCommand()
    async def check_allowed_structure_type_constraints(
        self,
        constrained_parameter: typing.Annotated[
            Native,
            sila.constraints.AllowedTypes(
                [
                    sila.data_types.Structure.create(
                        {"item": sila.data_types.Element("Item", "Item", "The actual item", sila.data_types.Integer)}
                    )
                ]
            ),
        ],
    ) -> None:
        """
        Test the AllowedTypes constraint with a Structure type on the Any type.

        Args:
          ConstrainedParameter: An Any type parameter with an AllowedTypes constraint that
            only allows a structure with an integer
        """



import collections.abc
import dataclasses
import datetime

from sila.framework.constraints import Set as SiLASet
from sila.framework.data_types import Date, Integer, Real, String, Time, Timestamp


@dataclasses.dataclass
class Set(SiLASet):
    """
    A constraint that enforces that a value is part of a defined set of values.

    Raises:
      ValueError: If the list of allowed values is empty.
    """

    values: collections.abc.Sequence[str | int | float | datetime.date | datetime.time | datetime.datetime]

    def __post_init__(self):
        if not self.values:
            msg = "Expected a non-empty list of values for Set constraint."
            raise ValueError(msg)

        values_type = type(self.values[0])

        if issubclass(values_type, str):
            self.values = [String(value) for value in self.values]
        elif issubclass(values_type, int):
            self.values = [Integer(value) for value in self.values]
        elif issubclass(values_type, float):
            self.values = [Real(value) for value in self.values]
        elif issubclass(values_type, datetime.datetime):
            self.values = [Timestamp.from_datetime(value) for value in self.values]
        elif issubclass(values_type, datetime.date):
            self.values = [Date.from_date(value) for value in self.values]
        elif issubclass(values_type, datetime.time):
            self.values = [Time.from_time(value) for value in self.values]
        else:
            msg = (
                "Expected type of str, int, float, datetime.date, datetime.time, datetime.datetime, "
                f"Instead received type '{values_type}'."
            )
            raise TypeError(msg)
        return super().__post_init__()



import asyncio
import logging

from unitelabs.cdk import Connector, run
from unitelabs.cdk.config import CloudServerConfig, ConnectorBaseConfig, SiLAServerConfig
from unitelabs.cdk.features.core.authorization_service import AuthorizationService
from unitelabs.cdk.features.test.any_type_test import AnyTypeTest
from unitelabs.cdk.features.test.authentication_test import AuthenticationTest
from unitelabs.cdk.features.test.basic_data_types_test import BasicDataTypesTest
from unitelabs.cdk.features.test.binary_transfer_test import BinaryTransferTest
from unitelabs.cdk.features.test.error_handling_test import ErrorHandlingTest
from unitelabs.cdk.features.test.list_data_type_test import ListDataTypeTest
from unitelabs.cdk.features.test.metadata_consumer_test import MetadataConsumerTest
from unitelabs.cdk.features.test.metadata_provider import MetadataProvider
from unitelabs.cdk.features.test.multi_client_test import MultiClientTest
from unitelabs.cdk.features.test.observable_command_test import ObservableCommandTest
from unitelabs.cdk.features.test.observable_property_test import ObservablePropertyTest
from unitelabs.cdk.features.test.parameter_constraints_test import ParameterConstraintsTest
from unitelabs.cdk.features.test.structure_data_type_test import StructureDataTypeTest
from unitelabs.cdk.features.test.unobservable_command_test import UnobservableCommandTest
from unitelabs.cdk.features.test.unobservable_property_test import UnobservablePropertyTest

from .features import AuthenticationService

logging.basicConfig(level=logging.INFO, force=True)


async def app_factory():
    app = Connector(
        ConnectorBaseConfig(
            sila_server=SiLAServerConfig(
                port=50052,
                tls=False,
                uuid="2be7e3fe-6a53-48fd-81d9-116f7cc5b59b",
                name="SiLA Python Interoperability Test Server",
                type="TestServer",
                version="0.1",
                description="This server aims to implement all features required for the SiLA Interoperability Test Suite",
                vendor_url="https://gitlab.com/SiLA2/sila_python",
            ),
            cloud_server_endpoint=CloudServerConfig(
                tls=False,
            ),
        )
    )
    app.register(AnyTypeTest())
    app.register(AuthenticationService())
    app.register(AuthenticationTest())
    app.register(AuthorizationService())
    app.register(BasicDataTypesTest())
    app.register(BinaryTransferTest())
    app.register(ErrorHandlingTest())
    app.register(ListDataTypeTest())
    app.register(MetadataConsumerTest())
    app.register(MetadataProvider())
    app.register(MultiClientTest())
    app.register(ObservableCommandTest())
    app.register(ObservablePropertyTest())
    app.register(ParameterConstraintsTest())
    app.register(StructureDataTypeTest())
    app.register(UnobservableCommandTest())
    app.register(UnobservablePropertyTest())

    yield app


if __name__ == "__main__":
    asyncio.run(run(app_factory))



from .labware_transfer_manipulator_controller_base import (
    LabwareTransferManipulatorControllerBase,
    InvalidCommandSequence,
    LabwareNotPicked,
    LabwareNotPlaced,
    PositionIndex,
    HandoverPosition,
)

__all__ = [
    "HandoverPosition",
    "InvalidCommandSequence",
    "LabwareNotPicked",
    "LabwareNotPlaced",
    "LabwareTransferManipulatorControllerBase",
    "PositionIndex",
]



# ruff: noqa: D205, D401, D415, E501

import typing_extensions as typing

from unitelabs.cdk import sila

from ...core.authorization_service import AccessToken


class AuthenticationTest(sila.Feature):
    """
    Contains commands that require authentication. A client should be able to obtain an Authorization Token through the Login command of the Authentication Service feature
    using the following credentials: username: 'test', password: 'test'
    """

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test")

    @sila.UnobservableCommand()
    async def requires_token(self, *, metadata: typing.Annotated[sila.Metadata, AccessToken]) -> None:
        """Requires an authorization token in order to be executed."""

    @sila.UnobservableCommand()
    def requires_token_for_binary_upload(
        self, binary_to_upload: bytes, *, metadata: typing.Annotated[sila.Metadata, AccessToken]
    ) -> None:
        """
        Requires an authorization token in order to be executed and to upload a binary parameter

        Args:
          BinaryToUpload: A binary that needs to be uploaded
        """



import enum

import pytest

from sila import String
from unitelabs.cdk.sila.common.feature import Feature
from unitelabs.cdk.sila.data_types.enum import Enum


class TestEnumDataType(enum.Enum):
    VALUE_A = enum.auto()
    VALUE_B = enum.auto()
    VALUE_C = enum.auto()


class TestWeirdCasingEnumDataType(enum.Enum):
    SiLA_value = enum.auto()


class TestFromNative:
    async def test_should_parse_from_native(self):
        # Create enum data type
        feature = Feature()
        data_type = Enum.create(TestEnumDataType)

        # Convert from native
        value = await data_type.from_native(feature.context, TestEnumDataType.VALUE_B)

        # Assert that the method returns the correct value
        assert value == data_type(value=String(value="value b"))


class TestToNative:
    async def test_should_parse_to_native(self):
        # Create enum data type
        feature = Feature()
        data_type = Enum.create(TestEnumDataType)
        value = data_type(value=String(value="value c"))

        # Convert to native
        native = await value.to_native(feature.context)

        # Assert that the method returns the correct value
        assert isinstance(native, TestEnumDataType)
        assert native == TestEnumDataType.VALUE_C

    async def test_should_parse_to_native_with_invalid_casing(self):
        # Create enum data type
        feature = Feature()
        data_type = Enum.create(TestWeirdCasingEnumDataType)
        value = data_type(value=String(value="sila value"))

        # Convert to native
        native = await value.to_native(feature.context)

        # Assert that the method returns the correct value
        assert isinstance(native, TestWeirdCasingEnumDataType)
        assert native == TestWeirdCasingEnumDataType.SiLA_value


class TestRoundTrip:
    @pytest.mark.parametrize(
        "enum,string,enum_value",
        [
            pytest.param(
                TestWeirdCasingEnumDataType, "sila value", TestWeirdCasingEnumDataType.SiLA_value, id="Invalid casing"
            ),
            pytest.param(TestEnumDataType, "value c", TestEnumDataType.VALUE_C, id="Valid casing"),
        ],
    )
    async def test_should_make_round_trip(self, enum, string, enum_value):
        feature = Feature()
        data_type = Enum.create(enum)
        value = data_type(value=String(value=string))

        # Convert to native
        native = await value.to_native(feature.context)

        # Assert that the method returns the correct value
        assert isinstance(native, enum)
        assert native == enum_value

        # convert back to sila
        sila_value = await value.from_native(feature.context, native)
        assert sila_value == value



import abc

import typing_extensions as typing

from unitelabs.cdk import sila, subscriptions

KELVIN = sila.constraints.Unit(label="K", components=[sila.constraints.UnitComponent(unit="Kelvin")])


class TemperatureNotReachable(Exception):
    """The ambient conditions prohibit the device from reaching the target temperature."""


class ControlInterrupted(Exception):
    """The control of temperature could not be finished as it has been interrupted by another 'Control Temperature' command."""  # noqa: E501


class TemperatureController(sila.Feature, metaclass=abc.ABCMeta):
    """
    This is a simple example of a generic Feature for controlling and retrieving the temperature.

    A new target temperature can be set anytime with the 'Control Temperature' Command.
    The temperature range has been limited to prevent major damages of a device.
    In case the first target temperature has not been reached, a ControlInterrupted Error should be thrown.
    """

    def __init__(self):
        super().__init__(
            originator="org.silastandard",
            category="examples",
            version="1.0",
            maturity_level="Verified",
        )

        self.current_temperature = 21.0
        self.target_temperature = self.current_temperature
        self.source = subscriptions.Publisher(source=self._change_temperature, interval=1)

    def _change_temperature(self) -> float:
        return self.current_temperature

    @sila.ObservableProperty()
    async def subscribe_current_temperature(self) -> sila.Stream[typing.Annotated[float, KELVIN]]:
        """Subscribe the current temperature as measured by the controller."""

        return self.source.subscribe()

    @sila.ObservableCommand()
    async def control_temperature(
        self,
        target_temperature: typing.Annotated[
            float, KELVIN, sila.constraints.MaximalInclusive(363.0), sila.constraints.MinimalExclusive(277.0)
        ],
        *,
        status: sila.Status,
    ) -> None:
        """
        Control the temperature gradually to a set target.

        It is RECOMMENDED to use an oscillation free control system.

        Args:
          TargetTemperature: The target temperature that the server will
            try to reach. Note that the command might be completed at a
            temperature that it evaluates to be close enough. If the
            temperature cannot be reached, a 'Temperature Not Reachable'
            error will be thrown.

        Raises:
          TemperatureNotReachable: The ambient conditions prohibit the
            device from reaching the target temperature.
          ControlInterrupted: The control of temperature could not be
            finished as it has been interrupted by another 'Control
            Temperature' command.
        """



import pytest

from unitelabs.cdk.sila.utils.name import to_display_name, to_identifier


class TestToDisplayName:
    def test_should_convert_empty_string(self):
        assert to_display_name("") == ""

    def test_should_convert_class_name(self):
        assert to_display_name("MyClassName") == "My Class Name"

    @pytest.mark.parametrize("value,name", [("MyClassUUID", "My Class UUID"), ("UUIDClass", "UUID Class")])
    def test_should_convert_class_name_with_consecutive_uppercase(self, value: str, name: str):
        assert to_display_name(value) == name

    def test_should_convert_variable_name(self):
        assert to_display_name("my_variable_name") == "My Variable Name"

    @pytest.mark.parametrize(
        "value,name", [("my_variable_UUID", "My Variable UUID"), ("UUID_variable", "UUID Variable")]
    )
    def test_should_convert_consecutive_uppercase(self, value: str, name: str):
        assert to_display_name(value) == name

    @pytest.mark.parametrize(
        "name",
        [
            ("My Class Name"),
            ("My Class UUID"),
            ("UUID Class"),
            ("My Variable Name"),
            ("My Variable UUID"),
            ("UUID Variable"),
        ],
    )
    def test_should_keep_display_name(self, name: str):
        assert to_display_name(name) == name


class TestToIdentifier:
    def test_should_convert_empty_string(self):
        assert to_identifier("") == ""

    @pytest.mark.parametrize(
        "name,identifier",
        [
            ("My Class Name", "MyClassName"),
            ("My Class UUID", "MyClassUUID"),
            ("UUID Class", "UUIDClass"),
            ("My Variable Name", "MyVariableName"),
            ("My Variable UUID", "MyVariableUUID"),
            ("UUID Variable", "UUIDVariable"),
        ],
    )
    def test_should_convert_display_name(self, name: str, identifier: str):
        assert to_identifier(name) == identifier

    @pytest.mark.parametrize(
        "name,identifier",
        [
            ("Héllo 123", "Hello123"),
            ("123 Start", "I123Start"),
            ("!Bad Chars#", "BadChars"),
        ],
    )
    def test_should_remove_special_characters(self, name: str, identifier: str):
        assert to_identifier(name) == identifier



import re

import packaging.version

SILA_SERVER_VERSION_PATTERN = r"(0|[1-9][0-9]*)\.(0|[1-9][0-9]*)(\.(0|[1-9][0-9]*))?(_[_a-zA-Z0-9]+)?"

SILA_SERVER_VERSION_RE = re.compile(SILA_SERVER_VERSION_PATTERN)


def parse_version(version: str) -> str:
    """Parse a PEP 440-compatible version into a valid SiLA 2 version."""

    try:
        v = packaging.version.parse(version)
        parts = [str(v.major), str(v.minor)]
        if v.micro != 0:
            parts.append(str(v.micro))
        sila_version = ".".join(parts)

        suffix_parts = []
        if v.pre:
            suffix_parts.append(f"{v.pre[0]}{v.pre[1]}")
        if v.post:
            suffix_parts.append(f"post{v.post}")
        if v.dev:
            suffix_parts.append(f"dev{v.dev}")
        if v.local:
            suffix_parts.append(v.local.replace(".", "_"))

        if suffix_parts:
            sila_version += "_" + "_".join(suffix_parts)

        return sila_version
    except packaging.version.InvalidVersion:
        if SILA_SERVER_VERSION_RE.fullmatch(version):
            return version

        msg = f"Invalid version format: '{version}'."
        raise ValueError(msg) from None



import dataclasses

import typing_extensions as typing

from ..common import Dataclass, Feature
from .custom import Custom


@dataclasses.dataclass
class CustomDataType(Dataclass):
    """
    A custom data type definition that can be reused in multiple places.

    Examples:
      Define a custom data type:
      >>> @dataclasses.dataclass
      ... class MyCustomDataType(sila.CustomDataType):
      ...   \"\"\"Describe what your data type is for.\"\"\"
      ...   param_a: str
      ...   param_b: int
      ...
      ... class MyFeature(sila.Feature):
      ...   @sila.UnobservableProperty()
      ...   async def my_property(self) -> MyCustomDataType:
      ...     \"\"\"Describe what your property does.\"\"\"
      ...     return MyCustomDataType(param_a="Hello, World!", param_b=42)
      ...
      ...   @sila.UnobservableCommand()
      ...   async def my_property(self, my_custom_data: MyCustomDataType) -> None:
      ...     \"\"\"
      ...     Describe what your command does.
      ...
      ...     Args:
      ...       my_custom_data: The custom data type to process.
      ...     \"\"\"
      ...     print(my_custom_data.param_a, my_custom_data.param_b)
    """

    @typing.override
    @classmethod
    def attach(cls, feature: Feature) -> type[Custom]:
        super().attach(feature)

        if cls._identifier in feature.data_type_definitions:
            return feature.data_type_definitions[cls._identifier]

        data_type = cls._infer_data_type(feature)
        cls._custom = Custom.create(
            identifier=cls._identifier,
            display_name=cls._name,
            description=cls._description,
            data_type=data_type,
            feature=feature,
        )
        cls._custom._class = cls

        return cls._custom



# ruff: noqa: E501

import dataclasses

from unitelabs.cdk import sila


@dataclasses.dataclass
class StringMetadata(sila.Metadatum):
    """A metadata consisting of a string. It affects the full "Metadata Consumer Test" feature."""

    string_metadata: str


@dataclasses.dataclass
class TwoIntegersMetadata(sila.Metadatum):
    """
    A metadata consisting of a structure with two integers. It affects only the command "Unpack Metadata" of the "Metadata Consumer Test" feature.

    Attributes:
      FirstInteger: The first integer
      SecondInteger: The second integer
    """

    first_integer: int
    second_integer: int


class MetadataProvider(sila.Feature):
    """This feature provides SiLA Client Metadata to the "Metadata Consumer Test" feature."""

    def __init__(self):
        super().__init__(originator="org.silastandard", category="test", metadata=[StringMetadata, TwoIntegersMetadata])



import dataclasses
import enum

import sila
from unitelabs.cdk.sila.common.feature import Feature
from unitelabs.cdk.sila.data_types.custom_data_type import CustomDataType
from unitelabs.cdk.sila.data_types.infer_data_type import infer


class TestCustom:
    def test_should_infer_custom_with_basic_type(self):
        feature = Feature()

        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            """Test Custom Data Type."""

            test_custom_data_type: str

        data_type = infer(TestCustomDataType, feature)

        assert issubclass(data_type, sila.Custom)
        assert data_type.identifier == "TestCustomDataType"
        assert data_type.display_name == "Test Custom Data Type"
        assert data_type.description == "Test Custom Data Type."
        assert data_type.data_type == sila.String

    def test_should_infer_custom_with_complex_type(self):
        feature = Feature()

        @dataclasses.dataclass
        class TestCustomDataType(CustomDataType):
            """
            Test Custom Data Type.

            Attributes:
              Value: A value.
              AnotherValue: Another value.
            """

            value: int
            another_value: float

        data_type = infer(TestCustomDataType, feature)

        assert issubclass(data_type, sila.Custom)
        assert data_type.identifier == "TestCustomDataType"
        assert data_type.display_name == "Test Custom Data Type"
        assert data_type.description == "Test Custom Data Type."
        assert issubclass(data_type.data_type, sila.Structure)
        assert data_type.data_type.elements == {
            "value": sila.Element("Value", "Value", "A value.", sila.Integer),
            "another_value": sila.Element("AnotherValue", "Another Value", "Another value.", sila.Real),
        }


class TestEnumDataType(enum.Enum):
    VALUE_A = enum.auto()
    VALUE_B = enum.auto()
    VALUE_C = enum.auto()


class TestEnum:
    def test_should_infer_enum(self):
        feature = Feature()

        data_type = infer(TestEnumDataType, feature)

        assert issubclass(data_type, sila.Constrained)
        assert data_type.data_type == sila.String
        assert data_type.constraints == [
            sila.Set([sila.String("value a"), sila.String("value b"), sila.String("value c")])
        ]

    def test_should_infer_list_of_enum(self):
        feature = Feature()

        data_type = infer(list[TestEnumDataType], feature)

        assert issubclass(data_type, sila.List)
        assert issubclass(data_type.data_type, sila.Constrained)
        assert data_type.data_type.constraints == [
            sila.Set([sila.String("value a"), sila.String("value b"), sila.String("value c")])
        ]

    def test_should_infer_enum_in_structure(self):
        feature = Feature()

        @dataclasses.dataclass
        class CustomEnumDataType:
            """
            Test Custom Data Type.

            Attributes:
              Value: A value.
            """

            value: TestEnumDataType

        data_type = infer(CustomEnumDataType, feature)

        assert issubclass(data_type, sila.Structure)
        assert issubclass(data_type.elements["value"].data_type, sila.Constrained)
        assert data_type.elements["value"].data_type.constraints[0].values == [
            sila.String("value a"),
            sila.String("value b"),
            sila.String("value c"),
        ]

    def test_should_infer_enum_in_custom(self):
        feature = Feature()

        @dataclasses.dataclass
        class CustomEnumDataType(CustomDataType):
            """
            Test Custom Data Type.

            Attributes:
              Value: A value.
            """

            value: TestEnumDataType

        data_type = infer(CustomEnumDataType, feature)

        assert issubclass(data_type, sila.Custom)
        assert issubclass(data_type.data_type, sila.Structure)
        assert issubclass(data_type.data_type.elements["value"].data_type, sila.Constrained)
        assert data_type.data_type.elements["value"].data_type.constraints[0].values == [
            sila.String("value a"),
            sila.String("value b"),
            sila.String("value c"),
        ]



from .authentication_service import AccessToken, AuthenticationFailed, AuthenticationService, InvalidAccessToken

__all__ = ["AccessToken", "AuthenticationFailed", "AuthenticationService", "InvalidAccessToken"]



import asyncio
import collections.abc
import functools
import importlib
import inspect
import itertools
import pathlib
import sys

import typing_extensions as typing


def coroutine(function: collections.abc.Callable) -> collections.abc.Callable:
    """Wrap click cli commands to run asynchronously."""

    @functools.wraps(function)
    def wrapper(*args, **kwargs):  # noqa: ANN202
        return asyncio.run(function(*args, **kwargs))

    return wrapper


def find_factory() -> str:
    """
    Search the factory method in the current working directory.

    Returns:
      The location of the factory method in the form 'module:name'.
    """

    sys_paths = [pathlib.Path(path) for path in sys.path]

    exception: Exception | None = None
    factories: list[str] = []
    for root, dirs, files in walk(pathlib.Path.cwd()):
        dirs[:] = [
            d
            for d in dirs
            if any(
                path in itertools.chain([pathlib.Path(root) / d], (pathlib.Path(root) / d).parents)
                for path in sys_paths
            )
        ]

        for name in files:
            filepath = pathlib.Path(root) / name

            if filepath.suffix != ".py":
                continue

            module_name = None
            for path in sys_paths:
                if path in filepath.parents:
                    rel_path = filepath.relative_to(path)
                    module_name = rel_path.parent.as_posix().replace("/", ".") + "." + rel_path.stem

            if module_name is None:
                continue

            try:
                module = importlib.import_module(module_name)
                if (
                    (factory := getattr(module, "create_app", None))
                    and (module := inspect.getmodule(factory))
                    and inspect.getsourcefile(factory) == str(filepath)
                ):
                    factories.append(f"{module.__name__}:create_app")
            except Exception as error:  # noqa: BLE001
                exception = exception or error

    if len(factories) > 1:
        raise ImportError(
            "Multiple `create_app` factories found. Please specify the correct one explicitly:\n"
            + "\n".join(f" - `{factory}`" for factory in factories)
        )

    if not factories:
        raise exception or ImportError(
            "No `create_app` factory found. Please check your application setup or provide the factory explicitly."
        )

    return factories[0]


# Polyfill `pathlib.Path.walk` (only available on python 3.12) from
# https://github.com/python/cpython/blob/main/Lib/pathlib/_abc.py#L651
def walk(self, top_down=True, on_error=None) -> typing.Iterable[tuple[str, list[str], list[str]]]:  # noqa
    """Walk the directory tree from this directory, similar to os.walk()."""
    paths = [self]
    while paths:
        path = paths.pop()
        if isinstance(path, tuple):
            yield path
            continue
        dirnames = []
        filenames = []
        if not top_down:
            paths.append((path, dirnames, filenames))
        try:
            for child in path.iterdir():
                try:
                    if child.is_dir():
                        if not top_down:
                            paths.append(child)
                        dirnames.append(child.name)
                    else:
                        filenames.append(child.name)
                except OSError:
                    filenames.append(child.name)
        except OSError as error:
            if on_error is not None:
                on_error(error)
            if not top_down:
                while not isinstance(paths.pop(), tuple):
                    pass
            continue
        if top_down:
            yield path, dirnames, filenames
            paths += [path.joinpath(d) for d in reversed(dirnames)]



import asyncio
import time
import unittest.mock

import pytest

from unitelabs.cdk.subscriptions import Subject, Subscription

ONE_OP_TIME = 0.01


class TestMaxsize:
    async def test_should_limit_queue_size(self):
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        for i in range(10):
            sub.update(f"value{i}")
        assert sub.size == 10

        with pytest.raises(asyncio.QueueFull):
            sub.update("value11")

        assert sub.size == 10


class TestSize:
    async def test_should_return_current_size(self):
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        assert sub.size == 0

        sub.update("value")
        assert sub.size == 1

        await sub.get()
        assert sub.size == 0


class TestUpdate:
    async def test_should_allow_none_value_in_subscription(self):
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        values_added = ["value", None]
        for value in values_added:
            sub.update(value)

        seen = []
        async for value in sub:
            if value is None:
                sub.cancel()

            seen.append(value)
        assert seen == values_added

    async def test_should_not_update_same_value(self):
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        sub.put_nowait = unittest.mock.Mock(wraps=sub.put_nowait)

        # update twice
        sub.update("value")
        sub.update("value")

        assert sub._value == "value"
        sub.put_nowait.assert_called_once_with("value")


class TestCancel:
    async def test_should_stop_waiting_for_queue_task_after_cancel(self):
        callback = unittest.mock.Mock()
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        sub.cancel()

        async for value in sub:
            # this is never executed because the iterator is immediately exited
            callback(value)

        await asyncio.sleep(ONE_OP_TIME * 2)
        callback.assert_not_called()

    async def test_should_stop_iterating_after_cancel_in_loop(self):
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        values = [f"value{i}" for i in range(1, 5)]
        for value in values:
            sub.update(value)

        seen = []
        async for value in sub:
            if value == values[2]:
                sub.cancel()

            seen.append(value)
        assert seen == values[:3]

    async def test_should_stop_iterating_after_cancel(self):
        start_time = time.perf_counter()
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        N_OPS = 5

        async def make_updates():
            for i in range(10):
                sub.update(f"value{i}")
                await asyncio.sleep(ONE_OP_TIME)

                if i == N_OPS - 1:
                    sub.cancel()

        update_task = asyncio.create_task(make_updates())

        seen = [value async for value in sub]
        assert seen == [f"value{i}" for i in range(N_OPS)]
        assert time.perf_counter() - start_time == pytest.approx(ONE_OP_TIME * (N_OPS + 1), rel=1e3)

        await update_task


class TestTerminate:
    async def test_should_remove_subscription(self):
        subject = Subject[str](maxsize=10)
        sub = subject.subscribe()
        assert sub in subject.subscribers
        sub.terminate()
        assert sub not in subject.subscribers


class TestAsyncIterator:
    async def test_should_clear_tasks_after_each_iteration(self):
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        for i in range(10):
            sub.update(f"value{i}")

        async for value in sub:
            assert "value" in value
            await asyncio.sleep(ONE_OP_TIME / 10)
            assert not sub._closed.is_set()
            if "9" in value:
                break

    async def test_should_raise_stop_iteration(self):
        # create a subscription and populate it with values
        sub = Subscription[str](10, parent=unittest.mock.Mock())

        task = asyncio.create_task(sub.__anext__())
        await asyncio.sleep(0)
        task.cancel()

        with pytest.raises(StopAsyncIteration):
            await task


class TestGet:
    async def test_should_get_new_value(self):
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        for i in range(10):
            msg = f"value{i}"
            sub.update(msg)
            assert await sub.get() == msg

    async def test_should_return_value_if_matches_predicate(self):
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        msgs = [f"value{i}" for i in range(9)]
        msgs.insert(5, "find me 5")
        for msg in msgs:
            sub.update(msg)

        assert await sub.get(lambda x: "find me" in x) == "find me 5"

    async def test_should_timeout_if_nothing_in_queue(self):
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        with pytest.raises(TimeoutError):
            await sub.get(timeout=0.05)

    async def test_should_timeout_if_nothing_matches_predicate(self):
        # create a subscription and populate it with values
        sub = Subscription[str](10, parent=unittest.mock.Mock())
        for value in [f"value {i}" for i in range(10)]:
            sub.update(value)

        with pytest.raises(TimeoutError):
            await sub.get(lambda x: "find me" in x, timeout=0.05)

    async def test_should_cycle_through_queued_items_until_match(self, create_task):
        sub = Subscription[str](10, parent=unittest.mock.Mock())

        # create a mock filter predicate to monitor how many times it was called
        def predicate(x: str) -> bool:
            return "5" in x

        mock_predicate = unittest.mock.Mock(wraps=predicate)

        # create background update task
        async def make_updates():
            for i in range(10):
                sub.update(f"value {i}")
                await asyncio.sleep(ONE_OP_TIME)

        update_task = next(create_task(make_updates()))

        assert await sub.get(mock_predicate) == "value 5"
        assert mock_predicate.call_count == 6

        # cleanup
        update_task.cancel()
        await asyncio.sleep(ONE_OP_TIME)
        assert update_task.done()



import collections.abc
import copy
import functools
import inspect
import weakref

import typing_extensions as typing
from sila.server import CommandExecution, Handler, Native, SiLAError, UndefinedExecutionError

from sila import Element, MetadataIdentifier

from ..data_types import to_sila
from ..metadata import Metadatum
from ..utils import parse_docstring, to_display_name, to_identifier
from .errors import define_error

if typing.TYPE_CHECKING:
    from .feature import Feature


class Decorator:
    """Base class for decorator based SiLA annotations."""

    def __init__(
        self,
        /,
        *,
        name: str | None = None,
        identifier: str | None = None,
        errors: collections.abc.Sequence[type[Exception]] | None = None,
        enabled: bool | collections.abc.Callable[..., bool] = True,
    ) -> None:
        self._identifier = identifier or ""
        self._name = name or ""
        self._description = ""
        self._enabled = enabled
        self._parameters: dict[str, Element] = {}
        self._responses: dict[str, Element] = {}
        self._intermediate_responses: dict[str, Element] = {}
        self._handler: Handler | None = None
        self._metadata: tuple[str, list[type[Metadatum]]] = ("", [])
        self._errors: list[type[Exception]] = list(errors or [])
        self._function: collections.abc.Callable = lambda: ...
        self._feature: Feature | None = None

    def __call__(self, function: collections.abc.Callable) -> collections.abc.Callable:
        """Call the decorator to wrap the given method."""

        self._function = weakref.proxy(function)
        setattr(function, "__handler", self)

        return function

    def is_enabled(self, feature: "Feature") -> bool:
        """Whether the handler is enabled or not."""

        if callable(self._enabled):
            return self._enabled(feature)

        return self._enabled

    def clone(self) -> "Decorator":
        """Create a copy of this decorator."""

        return copy.copy(self)

    def attach(self, feature: "Feature") -> bool:
        """
        Create and attach a handler to the `feature`.

        Args:
          feature: The `Feature` to which the handler will be attached.

        Returns:
          Whether the handler was attached or not.
        """

        if not self.is_enabled(feature):
            return False

        if self.__class__ is Decorator:
            docstring = parse_docstring(self._function, feature=feature)
            self._name = self._name or to_display_name(self._function.__name__)
            self._identifier = self._identifier or to_identifier(self._name)
            self._description = docstring.description

        self._feature = weakref.proxy(feature)

        return True

    async def execute(self, metadata: dict[MetadataIdentifier, Native], **parameters) -> Native:
        """
        Execute a given function with the provided keyword arguments.

        Args:
          metadata: Additional metadata sent from client to server.

        Returns:
          The result of the `function` execution.

        Raises:
          DefinedExecutionError: If the error type is in the list of defined errors.
          UndefinedExecutionError: If an unexpected error occurs during execution.
        """

        if not self._feature:
            raise RuntimeError

        try:
            function = self._with_metadata(self._function, metadata)
            function = self._with_parameters(function, parameters)

            return await self._execute(function)
        except SiLAError:
            raise
        except Exception as error:
            import traceback

            traceback.print_exc()

            error_name = error.__name__ if inspect.isclass(error) else error.__class__.__name__
            if any(error_name in e.__name__ for e in self._errors):
                raise define_error(error)(str(error)) from None

            msg = f"{error.__class__.__name__}: {error}"
            raise UndefinedExecutionError(msg) from error

    async def _execute(self, function: collections.abc.Callable) -> Native:
        responses = function()

        if inspect.isawaitable(responses):
            responses = await responses

        return to_sila(responses, self._responses)

    def _with_metadata(
        self, function: collections.abc.Callable, metadata: dict[MetadataIdentifier, Native]
    ) -> collections.abc.Callable:
        if self._metadata[0]:
            function = functools.partial(
                function,
                **{
                    self._metadata[0]: {
                        metadatum: metadatum.from_native(metadata[metadatum._metadatum.fully_qualified_identifier()])
                        for metadatum in self._metadata[1]
                        if metadatum._metadatum
                    }
                },
            )

        return function

    def _with_parameters(
        self,
        function: collections.abc.Callable,
        parameters: collections.abc.Mapping[str, Native | CommandExecution],
    ) -> collections.abc.Callable:
        if not self._feature:
            raise RuntimeError

        parameters = dict(parameters)
        command_execution = parameters.pop("command_execution", None)
        parameters = typing.cast(dict[str, Native], parameters)

        if command_execution and isinstance(command_execution, CommandExecution):
            from ..command import Intermediate, Status

            signature = inspect.signature(function)
            for param in signature.parameters.values():
                annotation = typing.get_origin(param.annotation) or param.annotation
                if annotation is Status:
                    function = functools.partial(function, **{param.name: Status(command_execution)})
                if annotation is Intermediate:
                    function = functools.partial(
                        function, **{param.name: Intermediate(command_execution, self._intermediate_responses)}
                    )

        return functools.partial(function, **parameters)
